"""
The 'sliceLen' function takes a slice object and an integer length and
returns the number of elements would be included if the slice were applied
to a sequence of that length.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  pass


def sliceLen(sliceObj: slice, length: int) -> int:
  """
  The 'sliceLen' function takes a slice object and an integer length and
  returns the number of elements would be included if the slice were applied
  to a sequence of that length.

  :param sliceObj: A slice object.
  :param length: An integer representing the length of the sequence.
  :return: The number of elements in the slice.
  """
  return len(range(*sliceObj.indices(length), ))



"""
The 'worktoy.utilities' module provides small, standalone utilities used
across the 'worktoy' library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._perm import perm
from ._slice_len import sliceLen
from ._type_cast import typeCast
from ._class_body_template import ClassBodyTemplate
from ._bipartite_matching import bipartiteMatching
from ._unpack import unpack
from ._maybe import maybe
from ._replace_flex import replaceFlex
from ._text_fmt import textFmt
from ._exception_info import ExceptionInfo
from ._string_list import stringList
from ._directory import Directory
from ._join_words import joinWords
from ._resolve_mro import resolveMRO
from ._word_wrap import wordWrap
from . import mathematics

__all__ = [
  'perm',
  'sliceLen',
  'typeCast',
  'ClassBodyTemplate',
  'bipartiteMatching',
  'unpack',
  'maybe',
  'replaceFlex',
  'textFmt',
  'ExceptionInfo',
  'stringList',
  'Directory',
  'joinWords',
  'resolveMRO',
  'wordWrap',
  'mathematics',
]



"""
The 'worktoy.utilities.mathematics' module provides mathematical utilities
used across the 'worktoy' library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._constants import pi, arcTan, atan2
from ._trig import sin, cos, tan, cosh, sinh, tanh
from ._exp_log import e, log, exp
from ._factorial import factorial

__all__ = [
  'e',
  'pi',
  'arcTan',
  'atan2',
  'log',
  'exp',
  'sin',
  'cos',
  'tan',
  'cosh',
  'sinh',
  'tanh',
  'factorial',
]



"""
The 'worktoy.ezdata' package provides the EZData dataclass.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen

from ._trust import trust
from ._ez_slot import EZSlot
from ._ez_desc import EZDesc
from ._ez_space_hook import EZSpaceHook
from ._ez_space import EZSpace
from ._ez_meta import EZMeta
from ._ez_data import EZData

__all__ = [
    'trust',
    'EZSlot',
    'EZDesc',
    'EZSpaceHook',
    'EZSpace',
    'EZMeta',
    'EZData',
]



"""
The 'worktoy.waitaminute' package provides the custom exceptions used
across the 'worktoy' library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from . import desc
from . import meta
from . import dispatch
from . import keenum
from . import ez
from ._attribute_error_factory import attributeErrorFactory
from ._type_exception import TypeException
from worktoy.waitaminute.desc._write_once_error import WriteOnceError
from ._variable_not_none import VariableNotNone
from ._unpack_exception import UnpackException
from ._path_syntax_exception import PathSyntaxException
from ._subclass_exception import SubclassException
from ._missing_variable import MissingVariable

___all__ = [
  'desc',
  'meta',
  'dispatch',
  'keenum',
  'ez',
  'attributeErrorFactory',
  'TypeException',
  'WriteOnceError',
  'VariableNotNone',
  'UnpackException',
  'PathSyntaxException',
  'SubclassException',
  'MissingVariable',
]



"""
KeeSpace provides the namespace class used by the 'worktoy.keenum' module.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..mcls import BaseSpace
from ..waitaminute.keenum import KeeDuplicate, KeeTypeException
from ..waitaminute.keenum import KeeNameConflict
from . import KeeSpaceHook, Kee

if TYPE_CHECKING:  # pragma: no cover
  from typing import Dict, TypeAlias

  Members: TypeAlias = Dict[str, Kee]
  Bases: TypeAlias = tuple[type, ...]


class KeeSpace(BaseSpace):
  """KeeSpace provides the namespace class used by the 'worktoy.keenum'
  module. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  keeSpaceHook = KeeSpaceHook()

  #  Private Variables
  __enumeration_members__ = None
  __member_type__ = None
  __num_list__ = None
  __null_value__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def addNum(self, name: str, member: Kee) -> None:
    """Adds the member to the enumeration dict. """
    if name in self.__enumeration_members__:
      oldMember = self.__enumeration_members__[name]
      raise KeeDuplicate(name, oldMember, member)
    try:
      oldName = member.name
    except AttributeError:
      member.name = name
    else:
      if oldName != name:
        raise KeeNameConflict(member, oldName, name)
    member.__num_index__ = len(self.__enumeration_members__)
    member = self.typeGuard(member)
    self.__enumeration_members__[name] = member

  def typeGuard(self, member: Kee) -> Kee:
    """Ensures that the member is an instance of KeeNum. """
    if self.__member_type__ is None:
      self.__member_type__ = member.__field_type__
      return member
    if self.__member_type__ is member.__field_type__:
      return member
    name, val, type_ = member.name, member.getValue(), self.__member_type__
    raise KeeTypeException(name, val, type_, )

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, mcls: type, name: str, bases: Bases, **kwargs) -> None:
    self.__enumeration_members__ = dict()
    super().__init__(mcls, name, bases, **kwargs)
    cls = type(self)

    for base in bases:
      try:
        space = getattr(base, '__namespace__')
      except AttributeError:
        continue
      else:
        try:
          memberType = space.__member_type__
        except AttributeError:
          continue
        else:
          self.__member_type__ = memberType
        try:
          members = getattr(space, '__enumeration_members__')
        except AttributeError:
          continue
        else:
          for i, (key, val) in enumerate(members.items()):
            if key in self.__enumeration_members__:
              oldMember = self.__enumeration_members__[key]
              raise KeeDuplicate(key, oldMember, val)
            self.addNum(key, val)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #



"""
The 'overload' function provides a decorator setting type signatures for
particular function overload. This overloading implementation requires
that the owning class is derived from 'BaseMeta' or a subclass of
'BaseMeta'. Other classes must use the 'Dispatcher' descriptor from
'worktoy.dispatch' instead.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..utilities import maybe, perm
from ..waitaminute import attributeErrorFactory
from . import TypeSig

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Callable, TypeAlias, Self, Iterator, Never

  Method: TypeAlias = Callable[..., Any]
  Decorator: TypeAlias = Callable[[Method], Self]


class overload:  # NOQA
  """
  Entry collected by LoadSpaceHook
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __sig_func_dict__ = None
  __next_sig__ = None
  __next_func__ = None
  __latest_func__ = None
  __fallback_func__ = None
  __finalizer_func__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getSigFuncDict(self) -> dict[TypeSig, Method]:
    return maybe(self.__sig_func_dict__, dict())

  def _getLatestFunc(self) -> Method:
    if self.__latest_func__ is None:
      raise RuntimeError
    return self.__latest_func__

  def isFallback(self) -> bool:
    """Check if the current overload is a fallback function."""
    return False if self.__fallback_func__ is None else True

  def getFallback(self, ) -> Method:
    """Get the fallback function for the overload."""
    return self.__fallback_func__

  def isFinalizer(self) -> bool:
    """Check if the current overload is a finalizer function."""
    return False if self.__finalizer_func__ is None else True

  def getFinalizer(self, ) -> Method:
    """Get the finalizer function for the overload."""
    return self.__finalizer_func__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _addSigFunc(self, sig: TypeSig, func: Method) -> None:
    existing = self._getSigFuncDict()
    existing[sig] = func
    self.__latest_func__ = func
    self.__sig_func_dict__ = existing

  def _extendLatest(self, sig: TypeSig) -> None:
    """Extend the latest function with a new signature."""
    self._addSigFunc(sig, self._getLatestFunc())

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __new__(cls, *types, **kwargs) -> Decorator:
    """Create a decorators that sets the type signature for the function."""

    if kwargs.get('_root', False):
      return super(overload, cls).__new__(cls)

    def decorator(func: Method) -> overload:
      if isinstance(func, cls):
        func._extendLatest(TypeSig(*types, ))
        return func
      self = super(overload, cls).__new__(cls)
      self._addSigFunc(TypeSig(*types, ), func)
      return self

    return decorator

  def __init__(self, *args, **kwargs) -> None:
    pass

  @classmethod
  def flex(cls, *types: type) -> Decorator:
    """Create a decorator that sets the type signature for the function."""

    def decorator(func: Method) -> Self:
      self = cls(_root=True)
      for p in perm(*types, ):
        self._addSigFunc(TypeSig(*p, ), func)
      return self

    return decorator

  @classmethod
  def fallback(cls, func) -> Self:
    """Create a decorator that sets the fallback function for the
    overload."""
    self = cls(_root=True)
    self.__fallback_func__ = func
    return self

  @classmethod
  def finalize(cls, func: Method) -> Self:
    """Create a decorator that sets the finalizer function for the
    overload."""
    self = cls(_root=True)
    self.__finalizer_func__ = func
    return self

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __getattr__(self, key: str) -> Any:
    funcs = [f for _, f in self._getSigFuncDict().items()]
    if funcs:
      func = funcs[0]
    elif self.isFallback():
      func = self.getFallback()
    else:
      raise attributeErrorFactory(self, key)
    try:
      value = getattr(func, key)
    except AttributeError:
      raise attributeErrorFactory(self, key)
    else:
      return value

  def __iter__(self, ) -> Iterator[tuple[TypeSig, Method]]:
    """Iterate over the signatures and functions in the overload."""
    yield from self._getSigFuncDict().items()

  if TYPE_CHECKING:  # pragma: no cover
    def __call__(self, func: Method) -> Never:
      """Linter friendly explicitly disabled call method. """



"""
EZDesc provides descriptors intended to provide class specific settings on
the EZMeta metaclass.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..core import Object
from ..utilities import maybe
from ..waitaminute import TypeException

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class EZDesc(Object):
  """
  EZDesc provides descriptors intended to provide class specific settings on
  the EZMeta metaclass.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # Fallback Variables
  __fallback_type__ = bool

  #  Private Variables
  __keyword_argument__ = None
  __default_value__ = None  # falls back to 'False'
  __value_type__ = None  # falls back to 'bool'

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getKwarg(self) -> Any:
    """
    Returns the keyword argument that controls the value of the
    descriptor. This must be set to a 'str' object when the descriptor is
    instantiated.
    """
    return self.__keyword_argument__

  def _getValueType(self) -> Any:
    """
    Returns the type of value that the descriptor is expected to hold. If
    a type other than 'bool' is required, it must be provided when
    instantiating the descriptor.
    """
    return maybe(self.__value_type__, self.__fallback_type__)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, key: str, *args) -> None:
    self.__keyword_argument__ = key
    type_, defVal, *_ = [*args, None, None]
    self.__value_type__ = type_
    self.__default_value__ = defVal

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_get__(self, *args, **_) -> Any:
    """
    This method is called when the descriptor is accessed on an instance.
    It should return the value of the descriptor for that instance.
    """
    kwargs = maybe(self.instance.__namespace__.__key_args__, dict())
    value = kwargs.get(self.__keyword_argument__, self.__default_value__)
    valueType = self._getValueType()
    if valueType is bool:
      return True if value else False
    if isinstance(value, valueType):
      return value
    raise TypeException('value', value, valueType, )



"""
AbstractKeeDesc provides an abstract baseclass for the descriptor classes
in the 'worktoy.keenum._kee_desc' module.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Type, TypeAlias

  from .. import KeeMeta

  Kee: TypeAlias = Type[KeeMeta]


class AbstractKeeDesc:
  """
  _KeeDesc provides a specialized descriptor class for the 'KeeMeta'
  class. It should be considered private to the 'worktoy.keenum' module and
  is not intended, nor fit, for general use.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __field_owner__ = None
  __field_name__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: Kee, name: str) -> None:
    """
    Set the owner and name of the descriptor when it is assigned to a class.
    This is called by Python when the descriptor is created.
    """
    self.__field_owner__ = owner
    self.__field_name__ = name

  def __get__(self, instance: KeeMeta, owner: Kee) -> Any:
    """
    Get the value of the descriptor. If the instance is None, return the
    owner class. Otherwise, return the value from the instance.
    """
    if instance is None:
      return self
    return self.__instance_get__(instance)

  @abstractmethod
  def __instance_get__(self, instance: KeeMeta) -> Any:
    """This method should be implemented by subclasses to specify
    particular functionality. """



"""
EZData leverages the 'worktoy' library to provide a dataclass.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import EZMeta
from ..mcls import BaseObject

if TYPE_CHECKING:  # pragma: no cover
  from typing import Callable, Iterator


def _root(callMeMaybe: Callable) -> Callable:
  """
  _root is a decorator that ensures the decorated function is called
  with the root class of EZData.
  """

  setattr(callMeMaybe, '__is_root__', True)
  return callMeMaybe


class EZData(BaseObject, metaclass=EZMeta):
  """
  EZData is a dataclass that provides a simple way to define data
  structures with validation and serialization capabilities.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @_root
  def __init__(self, *args, **kwargs) -> None:
    """This is just here for type checking purposes. The EZMeta control
    flow removes it with the auto-generated __init__ method."""

  @_root
  def __iter__(self, ) -> Iterator:
    """See documentation for __init__ above."""

  @_root
  def __len__(self, ) -> int:
    """See documentation for __init__ above."""

  @_root
  def __setitem__(self, *_) -> None:
    """See documentation for __init__ above."""

  @_root
  def __getitem__(self, *_) -> None:
    """See documentation for __init__ above."""

  @classmethod
  def __class_len__(cls, ) -> int:
    """Return the number of class variables."""
    return len(getattr(cls, '__slots__', ()))

  @classmethod
  def __class_iter__(cls, ) -> Iterator[str]:
    """Iterate over the class variables."""
    yield from getattr(cls, '__slots__', ())

  @classmethod
  def __class_contains__(cls, item: str) -> bool:
    """Check if the class contains the given item."""
    return item in getattr(cls, '__slots__', ())

  def __set_name__(self, owner: type, name: str) -> None:
    """Removes 'Object.__set_name__' which would set attributes on the
    instance. """



"""
TypeSig encapsulates type signatures for overloads
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..core.sentinels import THIS
from ..utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import TypeAlias, Self, Iterator, Union

  RawTypes: TypeAlias = tuple[type, ...]
  HASHABLE: TypeAlias = Union[type, int]


class TypeSig:
  """TypeSig encapsulates type signatures for overloads. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __raw_types__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getRawTypes(self) -> RawTypes:
    return self.__raw_types__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *rawTypes: type) -> None:
    self.__raw_types__ = rawTypes

  @classmethod
  def fromArgs(cls, *args, ) -> Self:
    """
    Create a TypeSig from the given arguments.
    """
    return cls(*[type(arg) for arg in args], )

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __iter__(self, ) -> Iterator[type]:
    yield from self._getRawTypes()

  def __hash__(self, ) -> int:
    return hash(self._getRawTypes())

  def __len__(self, ) -> int:
    return len(self._getRawTypes())

  def __contains__(self, type_: HASHABLE) -> bool:
    for rawType in self._getRawTypes():
      if rawType is type_:
        return True
    return False

  def __eq__(self, other: object) -> bool:
    if not isinstance(other, type(self)):
      return NotImplemented
    if len(self) != len(other):
      return False
    for this, that in zip(self, other):
      if this is not that:
        return False
    return True

  def __str__(self) -> str:
    """Returns a string representation of the type signature."""
    infoSpec = """%s object with %d types: %s"""
    typeStr = '[%s]' % ', '.join(str(t) for t in self)
    n = len(self)
    clsName = type(self).__name__
    return textFmt(infoSpec % (clsName, n, typeStr))

  def __repr__(self) -> str:
    """Returns code that would recreate the type signature."""
    infoSpec = """%s(%s)"""
    typeStr = ', '.join(repr(t) for t in self)
    return textFmt(infoSpec % (type(self).__name__, typeStr))

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def swapTHIS(self, thisType: type) -> None:
    newTypes = []
    for rawType in self._getRawTypes():
      if rawType is THIS:
        newTypes.append(thisType)
      else:
        newTypes.append(rawType)
    self.__raw_types__ = (*newTypes,)



"""EZSpace provides the namespace for the EZData class. """
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import EZSpaceHook
from ..mcls import BaseSpace

if TYPE_CHECKING:  # pragma: no cover
  pass


class EZSpace(BaseSpace):
  """
  EZSpace provides the namespace for the EZData class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Public Variables
  ezHook = EZSpaceHook()



"""
The 'worktoy.dispatch' package provides the overload functionality used
across the 'worktoy' library. It provides the central 'Dispatch' class
which facilitates mapping from type signatures to function objects.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._type_sig import TypeSig
from ._dispatcher import Dispatcher
from ._overload import overload, overload

__all__ = [
  'TypeSig',
  'Dispatcher',
  'overload',
  'overload',
]



"""
KeeMemberError provides a custom exception raised to indicate that an
object that is an instance of 'KeeNum' is not a member of a particular
enumeration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from ...keenum import KeeMeta, KeeNum


class KeeMemberError(KeyError):
  """
  KeeMemberError provides a custom exception raised to indicate that an
  object that is an instance of 'KeeNum' is not a member of a particular
  enumeration.
  """

  __slots__ = ('keenum', 'member')

  def __init__(self, keenum: KeeMeta, member: KeeNum) -> None:
    self.keenum, self.member = keenum, member

  def __str__(self) -> str:
    infoSpec = """KeeNum object: '%s' is not a member of the enumeration 
    '%s'!"""
    info = infoSpec % (self.member, self.keenum)
    from ...utilities import textFmt
    return textFmt(info, )

  __repr__ = __str__



"""
KeeNameConflict provides a custom exception raised to indicate that a
'KeeNum' enumeration received a 'Kee' member reservation already named,
but provided at a different name.
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any

  from ...keenum import Kee


class KeeNameConflict(ValueError):
  """
  KeeNameConflict provides a custom exception raised to indicate that a
  'KeeNum' enumeration received a 'Kee' member reservation already named,
  but provided at a different name.
  """

  member: Any
  oldName: str
  newName: str

  __slots__ = ('__kee_object__', '__existing_name__', '__new_name__',)

  def __init__(self, kee: Kee, oldName: str, newName: str) -> None:
    """Initialize the KeeNameConflict object."""
    self.member = kee
    self.oldName = oldName
    self.newName = newName
    ValueError.__init__(self, )

  def __str__(self, ) -> str:
    infoSpec = """Name conflict for Kee member object '%s': existing name 
    '%s' versus new name '%s'!"""
    keeName = getattr(self.member, '__name__', 'Unknown')
    info = infoSpec % (keeName, self.oldName, self.newName,)
    return info

  __repr__ = __str__



"""
FALLBACK provides a sentinel object indicating that an entry in a 'Dispatch'
object is a fallback entry. In particular, this is the case when a
'TypeSig' object contains only the 'FALLBACK' sentinel and no other type.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import Sentinel

if TYPE_CHECKING:  # pragma: no cover
  pass


class FALLBACK(Sentinel):
  """FALLBACK provides a sentinel object indicating that an entry in a
  'Dispatch' object is a fallback entry. In particular, this is the case
  when a 'TypeSig' object contains only the 'FALLBACK' sentinel and no
  other type."""



"""
KeeDuplicate is a custom exception raised to indicate that a KeeNum
class received a duplicate entry for an enumeration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from ...keenum import Kee


class KeeDuplicate(Exception):
  """
  KeeDuplicate is a custom exception raised to indicate that a KeeNum
  class received a duplicate entry for an enumeration.
  """

  __slots__ = ('name', 'oldMember', 'newMember',)

  def __init__(self, name: str, *members: Kee) -> None:
    """Initialize the KeeDuplicate object."""
    self.name = name
    self.oldMember = members[0]
    self.newMember = members[1]
    Exception.__init__(self, )

  def __str__(self, ) -> str:
    """Return the string representation of the KeeDuplicate object."""
    infoSpec = """Enumeration name '%s' already contains member: '%s', 
    but attempted to add duplicate member: '%s'!"""
    oldStr = str(self.oldMember)
    newStr = str(self.newMember)
    info = infoSpec % (self.name, oldStr, newStr,)
    return textFmt(info)

  __repr__ = __str__



"""
KeeFlag provides a singular enumeration in the KeeFlags class. Like the
'KeeNum' class, it is itself an enumeration, but where the 'KeeNum' class
provides only singular members, the 'KeeFlags' populate the enumeration
with the 'NULL' enumeration, the 'FULL' enumeration and all possible
combinations of the singular members.

Where the 'Kee' class allows the novel syntactic sugar of 'Kee[int](1)',
'KeeFlag' requires the owning 'KeeFlags' class to specify the type of the
'value' of each member. While duplicate 'name' attributes between members
are disallowed, duplicate 'value' attributes are allowed. Please note
however that when retrieving members by value, the member having the
lowest index is used.

Attributes (becomes attributes of the instances of the owning 'KeeFlags'
class):
- name: The name of the member. It is this name that is passed to the
'__set_name__' method during class creation.
- value: The value of the member. See note about value instantiation.
- 'index': The number of previously defined members when this member was
being defined (received '__set_name__' call). By default, 'KeeFlags'
auto-generates a member named 'NULL' with index 0.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..desc import Field
from ..utilities import textFmt, maybe
from ..waitaminute import MissingVariable, TypeException

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Tuple, Self, Iterator
  from ..keenum import KeeFlagsMeta


class KeeFlag:
  """
  KeeFlag provides a singular enumeration in the KeeFlags class. Like the
  'KeeNum' class, it is itself an enumeration, but where the 'KeeNum' class
  provides only singular members, the 'KeeFlags' populate the enumeration
  with the 'NULL' enumeration, the 'FULL' enumeration and all possible
  combinations of the singular members.

  Please note that the actual instances of this class are *not* the
  members of the enumeration. 'KeeFlags' manages a container of these.
  Then the members are created as wrappers on these instances.

  Where the 'Kee' class allows the novel syntactic sugar of 'Kee[int](1)',
  'KeeFlag' requires the owning 'KeeFlags' class to specify the type of the
  'value' of each member. While duplicate 'name' attributes between members
  are disallowed, duplicate 'value' attributes are allowed. Please note
  however that when retrieving members by value, the member having the
  lowest index is used.

  Attributes (becomes attributes of the instances of the owning 'KeeFlags'
  class):
  - name: The name of the member. It is this name that is passed to the
    '__set_name__' method during class creation.
  - value: The value of the member. See note about value instantiation.
  - 'index': The number of previously defined members when this member was
    being defined (received '__set_name__' call). By default, 'KeeFlags'
    auto-generates a member named 'NULL' with index 0. Please note that
    this value gets assigned by the metaclass machinery *before* the
    KeeFlags class is created.
  - 'args': Positional arguments passed to the constructor.
  - 'kwargs': Keyword arguments passed to the constructor.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Type hints for the benefit of linters
  value: Any
  lows: tuple
  highs: tuple
  names: tuple[str, ...]

  #  Class Variables

  #  Fallback Variables

  #  Private Variables
  __field_name__ = None
  __field_owner__ = None
  __member_index__ = None
  __member_name__ = None
  __pos_args__ = None
  __key_args__ = None

  #  Public Variables
  fieldName = Field()
  fieldOwner = Field()
  index = Field()
  name = Field()
  args = Field()
  kwargs = Field()

  #  Virtual Variables
  valueType = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @fieldName.GET
  def _getFieldName(self, ) -> str:
    if self.__field_name__ is None:
      raise MissingVariable(self, '__field_name__', str)
    if isinstance(self.__field_name__, str):
      return self.__field_name__
    raise TypeException('__field_name__', self.__field_name__, str)

  @fieldOwner.GET
  def _getFieldOwner(self, ) -> type:
    if self.__field_owner__ is None:
      from . import KeeFlagsMeta
      raise MissingVariable(self, '__field_owner__', KeeFlagsMeta)
    if isinstance(self.__field_owner__, type):
      return self.__field_owner__
    name, value = '__field_owner__', self.__field_owner__
    from . import KeeFlagsMeta
    raise TypeException(name, value, KeeFlagsMeta)

  @index.GET
  def _getIndex(self, ) -> int:
    if self.__member_index__ is None:
      raise MissingVariable(self, '__member_index__', int)
    if isinstance(self.__member_index__, int):
      return self.__member_index__
    raise TypeException('__member_index__', self.__member_index__, int)

  @name.GET
  def _getName(self, ) -> str:
    if self.__member_name__ is None:
      raise MissingVariable(self, '__member_name__', str)
    if isinstance(self.__member_name__, str):
      return self.__member_name__
    raise TypeException('__member_name__', self.__member_name__, str)

  @args.GET
  def _getArgs(self, ) -> Tuple[Any, ...]:
    return maybe(self.__pos_args__, ())

  @kwargs.GET
  def _getKwargs(self, ) -> dict[str, Any]:
    return maybe(self.__key_args__, {})

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: KeeFlagsMeta, name: str) -> None:
    self.__field_owner__ = owner
    self.__field_name__ = name

  def __str__(self, ) -> str:
    if self.__field_owner__ is None:
      return object.__str__(self)
    infoSpec = """%s.%s  [%d]"""
    clsName = self.fieldOwner.__name__
    return textFmt(infoSpec % (clsName, self.name, self.index))

  def __repr__(self, ) -> str:
    if self.__field_owner__ is None:
      return object.__repr__(self)
    infoSpec = """%s.%s [%d]: %s(%s)"""
    ownerName = self.fieldOwner.__name__
    fieldName = self.fieldName
    index = self.index
    clsName = type(self).__name__
    argStr = ', '.join(repr(a) for a in self.args)
    kwSpec = """%s=%s"""
    kwStr = ', '.join(kwSpec % (k, repr(v)) for k, v in self.kwargs.items())
    allArgs = []
    if argStr:
      allArgs.append(argStr)
    if kwStr:
      allArgs.append(kwStr)
    allArgStr = ', '.join(allArgs)
    info = infoSpec % (ownerName, fieldName, index, clsName, allArgStr)
    return textFmt(info)

  def __int__(self, ) -> int:
    return self._getIndex()

  def __iter__(self, ) -> Iterator[Self]:
    raise TypeError('KeeFlag is not iterable!')

  def __eq__(self, other: Self, **kwargs) -> bool:
    cls = type(self)
    selfOwner = self.__field_owner__
    otherOwner = other.__field_owner__
    if selfOwner != otherOwner:
      return NotImplemented
    return True if self.index == other.index else False

  def __hash__(self, ) -> int:
    return hash((self.__field_owner__, self.__field_name__, self.index))

  __index__ = _getIndex

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args: Any, **kwargs: Any) -> None:
    self.__pos_args__ = args
    self.__key_args__ = kwargs

  def clone(self, owner: type, index: int = None) -> Self:
    """Creates a clone of this KeeFlag for the specified owner."""
    cls = type(self)
    cloned = cls(*self.args, **self.kwargs)
    cloned.__field_owner__ = owner
    cloned.__field_name__ = self.__field_name__
    cloned.__member_index__ = maybe(index, self.__member_index__)
    cloned.__member_name__ = self.__member_name__
    return cloned

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #



"""
UnorderedEZException provides a custom exception raised when attempting to
sort an EZData subclass not supporting ordering. This is either because
the 'isOrdered' attribute is specifically set to False (default is True),
or if the subclass owns a field of a type that does not support ordering.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class UnorderedEZException(TypeError):
  """
  UnorderedEzException provides a custom exception raised when attempting to
  sort an EZData subclass not supporting ordering. This is either because
  the 'isOrdered' attribute is specifically set to False (default is True),
  or if the subclass owns a field of a type that does not support ordering.
  """

  __slots__ = ('className', 'fieldName', 'fieldType')

  def __init__(self, *args) -> None:
    """
    Initialize the UnorderedEzException exception.

    :param className: The name of the EZData subclass that is unordered.
    :param fieldName: Optional name of the field that does not support
    ordering.
    """
    cls, fName, fType, *_ = [*args, None, None]
    self.className = getattr(cls, '__name__', cls)
    self.fieldName, self.fieldType = None, None
    if fType is not None:
      self.fieldType = fType
      self.fieldName = fName
    TypeError.__init__(self, )

  def _isOrderedStr(self) -> str:
    """When 'isOrdered' is 'False'."""
    infoSpec = """Attempted to sort EZData subclass '%s' which has the 
    'order' class keyword set to 'False'. """
    info = infoSpec % self.className
    from worktoy.utilities import textFmt
    return textFmt(info, )

  def _fieldOrderedStr(self) -> str:
    """When a field does not support ordering."""
    infoSpec = """Attempted to sort EZData subclass '%s' which has the 
    field '%s' of type: '%s' that does not support ordering. """
    clsName = self.className
    fName = self.fieldName
    typeName = self.fieldType.__name__
    info = infoSpec % (clsName, fName, typeName)
    return textFmt(info, )

  def __str__(self) -> str:
    if self.fieldName is None or self.fieldType is None:
      return self._isOrderedStr()
    return self._fieldOrderedStr()

  __repr__ = __str__



"""The 'worktoy.work_io' module provides I/O functionalities. """
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._validate_existing_directory import validateExistingDirectory
from ._validate_existing_file import validateExistingFile
from ._validate_available_path import validateAvailablePath
from ._fid_gen import FidGen
from ._scrap_directory import scrapDirectory
from ._new_directory import newDirectory
from ._yeet_directory import yeetDirectory

__all__ = [
  'validateExistingDirectory',
  'validateExistingFile',
  'validateAvailablePath',
  'FidGen',
  'scrapDirectory',
  'newDirectory',
  'yeetDirectory',
]



"""
The 'worktoy.waitaminute.desc' package provides the custom exceptions
specific to the implementation of the descriptor protocol in the 'worktoy'
library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._descriptor_exception import DescriptorException
from ._access_error import AccessError
from ._protected_error import ProtectedError
from ._read_only_error import ReadOnlyError
from ._without_exception import WithoutException

__all__ = [
  'DescriptorException',
  'AccessError',
  'ReadOnlyError',
  'ProtectedError',
  'WithoutException',
]



"""
The 'worktoy.core' module provides the most primitive objects used by the
'worktoy' library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from . import sentinels
from ._instance_owner import ContextInstance, ContextOwner
from ._meta_type import MetaType
from ._object import Object

__all__ = [
  'sentinels',
  'ContextInstance',
  'ContextOwner',
  'MetaType',
  'Object',
]



"""
LoadSpaceHook collects DescLoad instances encountered in the class body
and creates an entry in the namespace.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...dispatch import overload, Dispatcher
from . import AbstractSpaceHook

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Type, TypeAlias

  Meta: TypeAlias = Type[type]


class LoadSpaceHook(AbstractSpaceHook):
  """
  LoadSpaceHook collects DescLoad instances encountered in the class body
  and creates an entry in the namespace.
  """

  def setItemPhase(self, key: str, val: Any, old: Any = None, ) -> bool:
    if not isinstance(val, overload):
      return False
    if val.isFallback():
      self.space.addFallback(key, val.getFallback())
      return True
    if val.isFinalizer():
      self.space.addFinalizer(key, val.getFinalizer())
      return True
    for sig, func in val:
      self.space.addOverload(key, sig, func)
    return True

  def postCompilePhase(self, compiledSpace) -> dict:
    """Populates the namespace with the collected DescLoad instances. """
    for name, sigFunc in self.space.getOverloads().items():
      dispatcher = Dispatcher()
      for sig, func in sigFunc.items():
        dispatcher.addSigFunc(sig, func)
      fallback = self.space.getFallbacks().get(name, None)
      if fallback is not None:
        dispatcher.setFallbackFunction(fallback)
      finalizer = self.space.getFinalizers().get(name, None)
      if finalizer is not None:
        dispatcher.setFinalizerFunction(finalizer)
      compiledSpace[name] = dispatcher
    return compiledSpace



"""
Members provides a descriptor returning the enumeration members defined in
this class and those inherited from the base class up to the root class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import AbstractKeeDesc

if TYPE_CHECKING:  # pragma: no cover
  from typing import Iterator
  from .. import KeeMeta, Kee


class Members(AbstractKeeDesc):
  """
  Members provides a descriptor returning the enumeration members defined in
  this class and those inherited from the base class up to the root class.
  """

  def __instance_get__(self, instance: KeeMeta) -> Iterator[Kee]:
    yield from getattr(instance, '__num_members__', [])



"""
KeeTypeException provides a custom exception raised to indicate member
with wrong type in an enumeration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class KeeTypeException(TypeError):
  """
  KeeTypeException provides a custom exception raised to indicate member
  with wrong type in an enumeration.
  """

  __slots__ = ('name', 'value', 'expectedTypes')

  def __init__(self, name: str, value: Any, *types) -> None:
    """Initialize the KeeTypeException object."""
    self.name = name
    self.value = value
    self.expectedTypes = types
    TypeError.__init__(self, )

  def __str__(self) -> str:
    """Return the string representation of the KeeTypeException object."""
    infoSpec = """KeeNum member '%s' has value '%s' of type '%s', but 
    expected type to be: '%s'!"""
    name = self.name
    typeNames = [t.__name__ for t in self.expectedTypes]
    from ...utilities import joinWords
    typeStr = joinWords(*["""'%s'""" % name for name in typeNames], )
    value = str(self.value)
    valueType = type(self.value).__name__
    info = infoSpec % (name, value, valueType, typeStr)
    return info

  __repr__ = __str__



"""
The 'worktoy.desc' module provides the base descriptor classes. This
module introduces a novel concept: descriptor-context.

When a descriptor is accessed through the owning class, the descriptor
object itself returns. When through an instance, the descriptor usually
performs the relevant accessor function as appropriate for the instance
received. This module expands this concept by introducing the
descriptor-context.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._alias import Alias
from ._field import Field
from ._attri_box import AttriBox
from ._fix_box import FixBox

__all__ = [
  'Alias',
  'Field',
  'AttriBox',
  'FixBox',
]



"""
perm: Efficient unique permutation generator.

This module provides the function `perm`, which yields all unique
permutations of its input arguments. The function works efficiently even
when the input contains duplicate elements, producing each unique ordering
exactly once.

Usage example:
  for p in perm('Tom', 'Tom', 'Harry'):
    print(p)
  # Outputs:
  # ('Tom', 'Tom', 'Harry')
  # ('Tom', 'Harry', 'Tom')
  # ('Harry', 'Tom', 'Tom')

The generator approach allows iteration over permutations without building
the entire result list in memory, making it suitable for large or memory-
constrained applications.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Iterator


def perm(*items: Any) -> Iterator[tuple[Any, ...]]:
  if not items:
    yield items
  dupes, dupe = [], lambda x: x in dupes or list.append(dupes, x)
  for j, t in enumerate(items):
    if not dupe(t):
      yield from [(t,) + p for p in perm(*items[:j], *items[j + 1:])]



"""
The 'trust' function decorates functions to indicate that they should be
ignored by the class creation process. It sets an attribute '__is_root__'
to True, which is treated as a signal to several space hook classes to
ignore the function. Decorated functions will not be available at runtime.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Callable


def trust(callMeMaybe: Callable) -> Callable:
  """Please note, that decorated functions will not be available at
  runtime. """
  setattr(callMeMaybe, '__is_root__', True)
  return callMeMaybe



"""
WILDCARD sentinel allows overloads to specify that a parameter can be any
value.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from . import Sentinel


class WILDCARD(Sentinel):
  """WILDCARD sentinel allows overloads to specify that a parameter can be
  any value."""



"""
Function is a special metaclass intended to represent objects that are
functions, builtin functions, or functions defined in class bodies,
but retrieved through instances or through the class itself.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...waitaminute.meta import IllegalInstantiation

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class _Root:
  pass


class _MetaFunction(type):
  """
  Function is a special metaclass intended to represent objects that are
  functions, builtin functions, or functions defined in class bodies,
  but retrieved through instances or through the class itself.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  __func_types__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def _createFunctionTypes(mcls, ) -> None:
    """
    Creates and caches the list of types that are considered function
    types.
    """

    def foo() -> None:
      """If you just put 'pass', it is not considered covered lmao"""

    class Bar:
      @classmethod
      def baz(cls) -> None:
        """If you just put 'pass', it is not considered covered lmao"""

      @staticmethod
      def qux() -> None:
        """If you just put 'pass', it is not considered covered lmao"""

      def quux(self) -> None:
        """If you just put 'pass', it is not considered covered lmao"""

    mcls.__func_types__ = [
      type(lambda: None),
      type(print),
      type(foo),
      type(Bar.baz),
      type(Bar.qux),
      type(Bar.quux),
      type(Bar().baz),
      type(Bar().qux),
      type(Bar().quux),
    ]

  @classmethod
  def _getFuncTypes(mcls, **kwargs) -> list[type]:
    """
    Returns a list of the types that are considered function types.
    """
    if mcls.__func_types__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      mcls._createFunctionTypes()
      return mcls._getFuncTypes(_recursion=True)
    return mcls.__func_types__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instancecheck__(cls, instance: object) -> bool:
    """
    This method validates as instances functions, builtin functions,
    lambda or class body defined functions.
    """
    if not callable(instance):
      return False

    for type_ in cls._getFuncTypes():
      if isinstance(instance, type_):
        return True
    return False

  def __subclasscheck__(cls, subclass: Any) -> bool:
    """
    This method validates as subclasses any class that has at least one
    callable attribute that is a function, builtin function, lambda or
    class body defined function.
    """
    try:
      _ = issubclass(subclass, type)
    except TypeError as typeError:
      raise typeError
    funcTypes = cls._getFuncTypes()
    for type_ in funcTypes:
      if issubclass(subclass, type_):
        return True
    return False

  def __new__(mcls, name: str, bases: tuple, space: dict, **kwargs) -> Any:
    """
    Prevents instantiation of the metaclass.
    """
    if _Root in bases:
      return super().__new__(mcls, name, (), space, **kwargs)

    raise IllegalInstantiation(mcls)


class Function(_Root, metaclass=_MetaFunction, ):
  """
  Function is a special metaclass intended to represent objects that are
  functions, builtin functions, or functions defined in class bodies,
  but retrieved through instances or through the class itself.
  """
  pass



"""
TypeCast encapsulates the logic for instantiating a type from arguments.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


def typeCast(target: type, arg: Any) -> Any:
  """
  Casts the given argument to the specified type, with the following
  special exceptions made for particular target types:
  - If the target is 'str', the given argument must be a 'str' object or
  one of 'bytes' or 'bytearray' that can be decoded to a 'str'. Otherwise,
  raises TypeCastException.
  - If the target is 'bool', the given argument must be one of the four
  following objects: 0, 1, True, or False. Otherwise, raises
  TypeCastException.
  - Target 'int' accepts args of the following types:
    - 'int' (returns the same value)
    - 'float' returns int(arg) if float.is_integer(arg), otherwise raises
    TypeCastException.
    - 'complex' returns int(arg.real) if arg.imag is 0 and
    float.is_integer(arg.real), otherwise raises TypeCastException.
    - 'str' uses try, except, else flow attempting to cast the string to
    'int' with int(arg). If this raises any exception, TypeCastException
    raises from that exception. Else the cast value is returned.
  - Target 'float' accepts args of the following types:
    - 'float' (returns the same value)
    - 'int' returns float(arg) for every integer arg.
    - 'complex' returns float(arg.real) if arg.imag is 0, otherwise raises
    TypeCastException.
    - 'str' uses try, except, else flow attempting to cast the string to
    'float' with float(arg). If this raises any exception, TypeCastException
    raises from that exception. Else the cast value is returned.
  - Target 'complex' accepts args of the following types:
    - 'complex' (returns the same value)
    - 'int' returns float(arg) + 0j.
    - 'float' returns float(arg) + 0j.
    - 'str' uses try, except, else flow attempting to cast the string to
    'complex' with complex(arg). If this raises any exception,
    TypeCastException raises from that exception. Else the cast value is
    returned.
  - Target 'list' accepts args of the following types:
    - 'list' (returns the same value)
    - 'tuple' returns list(arg).
    - 'set' returns list(arg).
    - 'frozenset' returns list(arg).
    - 'dict' returns list(dict.items(arg)).
  - Target 'tuple' accepts the exact same types as 'list' but returns a
  tuple instead of a list.
  - Target 'set' accepts the exact same types as 'list' but returns a set
  instead of a list.
  - Target 'frozenset' accepts the exact same types as 'list' but returns a
  frozenset instead of a list.
  - Target 'dict' accepts only other dictionary-like objects. This is
  achieved by a try, except, else flow attempting to: {**arg,}. If any
  exception is caught, TypeCastException raises from that exception.
  - Target 'type' accepts only classes, as defined by isinstance(arg, type).
  - Target 'Func' accepts nothing. Not even other functions.
  - For all other target types, if the argument is not an instance of
  the target type, it is attempted to be cast to the target type using
  the target type's constructor. If this raises any exception, a
  TypeCastException is raised from that exception.
  """
  if isinstance(arg, target):
    return arg
  if target is type:
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg)
  if target is str:
    if isinstance(arg, (bytes, bytearray)):
      try:
        return arg.decode('utf-8')
      except Exception as exception:
        from worktoy.waitaminute.dispatch import TypeCastException
        raise TypeCastException(target, arg) from exception
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg)
  if target is bool:
    if arg in (True, False, 0, 1):
      return bool(arg)
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg)
  if target is int:
    if isinstance(arg, float):
      if float.is_integer(arg):
        return int(arg)
      from worktoy.waitaminute.dispatch import TypeCastException
      raise TypeCastException(target, arg)
    if isinstance(arg, complex):
      if arg.imag == 0 and float.is_integer(arg.real):
        return int(arg.real)
      from worktoy.waitaminute.dispatch import TypeCastException
      raise TypeCastException(target, arg)
    if isinstance(arg, str):
      try:
        return int(arg)
      except Exception as exception:
        from worktoy.waitaminute.dispatch import TypeCastException
        raise TypeCastException(target, arg) from exception
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg)
  if target is float:
    if isinstance(arg, int):
      return float(arg)
    if isinstance(arg, complex):
      if arg.imag == 0:
        return float(arg.real)
      from worktoy.waitaminute.dispatch import TypeCastException
      raise TypeCastException(target, arg)
    if isinstance(arg, str):
      try:
        return float(arg)
      except Exception as exception:
        from worktoy.waitaminute.dispatch import TypeCastException
        raise TypeCastException(target, arg) from exception
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg)
  if target is complex:
    if isinstance(arg, (int, float)):
      return float(arg) + 0j
    if isinstance(arg, str):
      try:
        return complex(arg)
      except Exception as exception:
        from worktoy.waitaminute.dispatch import TypeCastException
        raise TypeCastException(target, arg) from exception
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg)
  if target is list:
    if isinstance(arg, (tuple, set, frozenset)):
      return list(arg)
    if isinstance(arg, dict):
      return [(k, v) for k, v in arg.items()]
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg)
  if target is tuple:
    if isinstance(arg, (list, set, frozenset)):
      return tuple(arg)
    if isinstance(arg, dict):
      return tuple((k, v) for k, v in arg.items())
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg)
  if target is set:
    if isinstance(arg, (list, tuple, frozenset)):
      return set(arg)
    if isinstance(arg, dict):
      return set((k, v) for k, v in arg.items())
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg)
  if target is frozenset:
    if isinstance(arg, (list, tuple, set)):
      return frozenset(arg)
    if isinstance(arg, dict):
      return frozenset((k, v) for k, v in arg.items())
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg)
  if target is dict:
    try:
      return {**arg}
    except Exception as exception:
      from worktoy.waitaminute.dispatch import TypeCastException
      raise TypeCastException(target, arg) from exception
  try:
    return target(arg)
  except Exception as exception:
    from worktoy.waitaminute.dispatch import TypeCastException
    raise TypeCastException(target, arg) from exception



"""
PLACE_HOLDER provides a sentinel object used to indicate that a value has
not yet been set. Intended for use in 'write-once' implementations. The
sentinel object should *not* be returned from any '__get__'
implementation, instead it should use fallback, create new object or raise
'AttributeError'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from worktoy.core.sentinels import Sentinel

if TYPE_CHECKING:  # pragma: no cover
  pass


class PLACE_HOLDER(Sentinel):
  """
  PLACE_HOLDER provides a sentinel object used to indicate that a value has
  not yet been set. Intended for use in 'write-once' implementations. The
  sentinel object should *not* be returned from any '__get__'
  implementation, instead it should use fallback, create new object or raise
  'AttributeError'.
  """



"""
FixBox subclasses 'AttriBox' and provides a descriptor for immutable
attributes. The only change reimplemented is that setting is only allowed
when no value has been set before.
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from worktoy.core import Object
from worktoy.desc import AttriBox
from worktoy.waitaminute import WriteOnceError

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class FixBox(AttriBox):
  """
  FixBox subclasses 'AttriBox' and provides a descriptor for immutable
  attributes. The only change reimplemented is that setting is only allowed
  when no value has been set before.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _inspectValue(self, ) -> Any:
    """
    Retrieves the value without invoking side effects.
    """
    pvtName = self.getPrivateName()
    try:
      value = object.__getattribute__(self.instance, pvtName)
    except AttributeError:
      return None
    else:
      return value

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_set__(self, value: Any, *args, **kwargs) -> None:
    oldValue = self._inspectValue()
    if oldValue is not None:
      raise WriteOnceError(self, oldValue, value)
    AttriBox.__instance_set__(self, value, *args, **kwargs)

  def __instance_delete__(self, *args, **kwargs) -> None:
    Object.__instance_delete__(self, *args, **kwargs)



"""The 'worktoy.waitaminute.keenum' module provides custom exceptions used
by the 'worktoy.keenum' module."""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._kee_name_conflict import KeeNameConflict
from ._kee_case_exception import KeeCaseException
from ._kee_duplicate import KeeDuplicate
from ._kee_flag_duplicate import KeeFlagDuplicate
from ._kee_member_error import KeeMemberError
from ._kee_index_error import KeeIndexError
from ._kee_name_error import KeeNameError
from ._kee_value_error import KeeValueError
from ._kee_type_exception import KeeTypeException
from ._kee_write_once_error import KeeWriteOnceError

__all__ = [
  'KeeNameConflict',
  'KeeCaseException',
  'KeeDuplicate',
  'KeeFlagDuplicate',
  'KeeMemberError',
  'KeeIndexError',
  'KeeNameError',
  'KeeValueError',
  'KeeTypeException',
  'KeeWriteOnceError',
]



"""
EZMultipleInheritance provides a custom exception raised when any attempt
is made to create a class that inherits from multiple subclasses of EZData.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import joinWords

if TYPE_CHECKING:  # pragma: no cover
  from typing import Type
  from worktoy.ezdata import EZData


class EZMultipleInheritance(TypeError):
  """
  EZMultipleInheritance provides a custom exception raised when any attempt
  is made to create a class that inherits from multiple subclasses of EZData.
  """

  __slots__ = ('name', 'bases')

  def __init__(self, name: str, *bases: Type[EZData]) -> None:
    """
    Initialize the EZMultipleInheritance exception.

    :param name: The name of the class being created.
    :param bases: The base classes that are being inherited from.
    """
    self.name = name
    self.bases = bases
    TypeError.__init__(self, )

  def __str__(self) -> str:
    infoSpec = """Attempted to create class '%s' with multiple EZData 
    subclasses: '%s'. """
    baseStr = joinWords(*[b for b in self.bases])
    return infoSpec % (self.name, baseStr)

  __repr__ = __str__



"""
ExceptionInfo provides a context manager for capturing and inspecting
exceptions. The motivating use case is when debugging scenarios where in
the testing framework, and expected exception fails to raise or does
raise, but is of unexpected 'Exception' subclass.
"""
#  AGPL-3.0 license
#  Copyright (c) 2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Type, Self, Optional, Any


class _Exc:
  """
  Private descriptor exposing exception class name for exception
  attributes. Falls back to 'No Exception' when absent.
  """

  __slot_name__ = None
  __field_name__ = None
  __field_owner__ = None

  def __init__(self, name: str) -> None:
    self.__slot_name__ = name

  def __set_name__(self, owner: type, name: str) -> None:
    self.__field_name__ = name
    self.__field_owner__ = owner

  def __get__(self, instance: Any, owner: type) -> Any:
    if instance is None:
      return self
    excAttr = getattr(instance, self.__slot_name__, )
    if isinstance(excAttr, type):
      return excAttr
    if isinstance(excAttr, Exception):
      return type(excAttr)
    return 'No Exception'


class _ExcName(_Exc):
  def __get__(self, instance: Any, owner: type) -> Any:
    if instance is None:
      return self
    excAttr = getattr(instance, self.__slot_name__, )
    if isinstance(excAttr, type):
      return excAttr.__name__
    if isinstance(excAttr, Exception):
      return type(excAttr).__name__
    return 'No Exception'


class _ExcObject(_Exc):
  def __get__(self, instance: Any, owner: type) -> Any:
    if instance is None:
      return self
    excAttr = getattr(instance, self.__slot_name__, )
    if isinstance(excAttr, BaseException):
      return excAttr
    return None


class ExceptionInfo:
  """
  ExceptionInfo provides a context manager for capturing and inspecting
  exceptions. The motivating use case is when debugging scenarios where in
  the testing framework, and expected exception fails to raise or does
  raise, but is of unexpected 'Exception' subclass.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Annotations
  __expected_exception__: Optional[Type[Exception]]
  __actual_exception__: Optional[Exception]
  report: str

  #  Public Variables
  expectedExcType = _Exc('__expected_exception__')
  actualExcType = _Exc('__actual_exception__')
  actualException = _ExcObject('__actual_exception__')
  expectedName = _ExcName('__expected_exception__')
  actualName = _ExcName('__actual_exception__')

  __slots__ = ('__expected_exception__', '__actual_exception__', 'report',)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, expExc: Type[Exception] = None, ) -> None:
    if isinstance(expExc, (type, BaseException)):
      self.__expected_exception__ = expExc
    else:
      self.__expected_exception__ = None
    self.__actual_exception__ = None
    self.report = ''

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __enter__(self, ) -> Self:
    return self

  def __exit__(self, _, excValue: Exception, __) -> bool:
    self.__actual_exception__ = excValue
    if not isinstance(excValue, Exception):
      if isinstance(excValue, BaseException):
        self.report = self._handleBaseException(excValue)
        return False  # BaseException must *always* propagate
    if excValue is None:
      self.report = self._handleNoException()
      return True
    if self.__expected_exception__ is None:
      return False  # Did not expect an exception, let it propagate
    if type(excValue) is self.expectedExcType:
      self.report = self._handleExpectedException()
      return True
    if isinstance(excValue, self.expectedExcType):
      self.report = self._handleSubclassException()
      return True
    self.report = self._handleUnexpectedException()
    return True

  def __str__(self) -> str:
    clsName = type(self).__name__
    infoSpec = """<%s expected='%s' actual='%s'>"""
    info = infoSpec % (clsName, self.expectedName, self.actualName)
    return textFmt(info)

  def __repr__(self, ) -> str:
    if self.__expected_exception__ is None:
      return textFmt("""%s()""" % (type(self).__name__,))
    infoSpec = """%s(%s)"""
    clsName = type(self).__name__
    info = infoSpec % (clsName, self.expectedName)
    return textFmt(info)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @staticmethod
  def _handleBaseException(baseExp: BaseException) -> str:
    infoSpec = """Received BaseException of type '%s', which is not a 
    subclass of 'Exception'. These exceptions must always propagate!"""
    excType = type(baseExp).__name__
    info = infoSpec % (excType,)
    return textFmt(info)

  def _handleNoException(self) -> str:
    infoSpec = """Expected '%s', but no exception was raised!"""
    excType = self.expectedName
    info = infoSpec % (excType,)
    return textFmt(info)

  def _handleSubclassException(self, ) -> str:
    infoSpec = """Expected '%s' and received exception of type: 
    '%s', a subclass of '%s': '%s'"""
    expName = self.expectedName
    excStr = str(self.actualException)
    info = infoSpec % (expName, self.actualName, expName, excStr)
    return textFmt(info)

  def _handleUnexpectedException(self, ) -> str:
    infoSpec = """Expected '%s', but received exception of type '%s' 
    instead: %s"""
    excStr = str(self.actualException)
    info = infoSpec % (self.expectedName, self.actualName, excStr,)
    return textFmt(info)

  def _handleExpectedException(self, ) -> str:
    infoSpec = """Caught '%s' as expected: %s"""
    info = infoSpec % (self.expectedName, str(self.actualException))
    return textFmt(info)



"""KeeFlagsMeta provides the metaclass for KeeFlags."""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..desc import Field
from ..mcls import BaseMeta
from ..utilities import textFmt, maybe
from . import KeeFlag
from . import KeeFlagsSpace as KFSpace

if TYPE_CHECKING:  # pragma: no cover
  from typing import TypeAlias, Self, Any, Iterator

  Bases: TypeAlias = tuple[type, ...]

  from . import KeeFlags


class KeeFlagsMeta(BaseMeta):
  """
  KeeFlagsMeta is the metaclass for KeeFlags, providing additional
  functionality for handling flags.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  __kee_class__ = None
  __kee_bases__ = None

  #  Private Variables
  __kee_flags__ = None
  __kee_members__ = None
  __allow_instantiation__ = None

  #  Public Variables
  flags = Field()
  memberList = Field()
  memberDict = Field()
  valueType = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @flags.GET
  def _getFlags(cls) -> list[KeeFlag]:
    return [flag for flag in cls.getKeeFlags().values()]

  @valueType.GET
  def _getValueType(cls) -> type:
    nullMember = cls.NULL
    return type(nullMember.value)

  @memberList.GET
  def _getMemberList(cls) -> list[KeeFlags]:
    return cls.__member_list__

  @memberDict.GET
  def _getMemberDict(cls) -> dict[str, KeeFlags]:
    return cls.__member_dict__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __subclasscheck__(cls, other: type, **kwargs) -> bool:
    if cls is other:
      return True
    if other is type(cls).__kee_class__:
      return False
    if not isinstance(other, type):
      raise TypeError('issubclass() arg 1 must be a class')
    otherKeeBases = getattr(other, '__kee_bases__', None)
    if otherKeeBases is None:
      return False
    out = None
    for base in otherKeeBases:
      if base is cls or cls.__subclasscheck__(base):
        return True
    return False

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kw) -> KFSpace:
    """Replaces the KeeSpace with KeeFlagsSpace"""
    bases = (*[b for b in bases if b.__name__ != '_InitSub'],)
    return KFSpace(mcls, name, bases, **kw)

  def __new__(mcls, name: str, bases: Bases, space: KFSpace, **kw) -> Self:
    """
    Creates a new instance of the class.
    This method is called when the class is created.
    """
    cls = BaseMeta.__new__(mcls, name, bases, space, **kw)
    if name == 'KeeFlags':
      setattr(cls, '__kee_bases__', bases)
      setattr(mcls, '__kee_class__', cls)
      return cls
    setattr(cls, '__kee_bases__', bases)
    cls.__allow_instantiation__ = True
    for flag in cls.flags:
      flag.__set_name__(cls, flag.name)
    memberList = []
    memberDict = dict()
    n = 2 ** len(cls.flags)
    for i in range(n):
      member = cls(i, )
      setattr(member, '__field_owner__', cls)
      setattr(member, '__field_name__', member.name)
      setattr(cls, member.name, member)
      memberList.append(member)
      memberDict[member.names] = member
    cls.__member_list__ = memberList
    cls.__member_dict__ = memberDict
    cls.__allow_instantiation__ = False
    customMRO = [cls, *bases, *cls.__mro__[1:]]
    valueGetter = None
    valueGetters = []
    for obj in customMRO:
      valueGetter = maybe(obj.__dict__.get('_getValue'), valueGetter)
      if valueGetter is None:
        continue
      if obj is mcls.__kee_class__:
        continue
      if valueGetter is getattr(mcls.__kee_class__, '_getValue'):
        continue
      valueGetters.append(valueGetter)
    valueGetters.append(getattr(mcls.__kee_class__, '_getValue'))
    setattr(cls, '_getValue', valueGetters[0])
    return cls

  def __call__(cls, *args, **kwargs) -> Any:
    if getattr(cls, '__allow_instantiation__', False):
      return BaseMeta.__call__(cls, *args, **kwargs)
    return cls._resolveMember(*args, **kwargs)

  def __len__(cls) -> int:
    return len(cls.memberList)

  def __iter__(cls, ) -> Iterator[KeeFlags]:
    yield from cls.memberList

  def __contains__(cls, identifier: Any) -> bool:
    try:
      _ = cls._resolveMember(identifier)
    except (IndexError, KeyError, ValueError, TypeError):
      return False
    else:
      return True

  def __getitem__(cls, identifier: Any) -> KeeFlags:
    return cls._resolveMember(identifier)

  def __eq__(cls, other: Any) -> bool:
    try:
      otherHash = hash(other)
    except TypeError as typeError:
      if 'hashable type' in str(typeError):
        return NotImplemented
      raise typeError
    else:
      return NotImplemented if otherHash == hash(cls) else False

  def __hash__(cls, ) -> int:
    return hash((cls.__name__, cls.__module__,))

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _resolveIndex(cls, index: int) -> KeeFlags:
    for member in cls:
      if member.index == index:
        return member
    infoSpec = """KeeFlags class '%s' has no member at index: '%d'!"""
    info = infoSpec % (cls.__name__, index,)
    raise IndexError(textFmt(info))

  def _resolveName(cls, name: str) -> KeeFlags:
    identifier = frozenset(name.upper().split('_'))
    for member in cls:
      if member.name == name or member.names == identifier:
        return member
    infoSpec = """KeeFlags class '%s' has no member with name: '%s'!"""
    info = infoSpec % (cls.__name__, name,)
    raise KeyError(textFmt(info))

  def _resolveNames(cls, *names: str) -> KeeFlags:
    return cls.memberDict[frozenset(names)]

  def _resolveValue(cls, value: Any) -> KeeFlags:
    for member in cls:
      if member.value == value:
        return member
    infoSpec = """KeeFlags class '%s' has no member with value: '%s'!"""
    info = infoSpec % (cls.__name__, value,)
    raise ValueError(textFmt(info))

  def _resolveMember(cls, *identifier: Any, **kwargs) -> KeeFlags:
    if not identifier:
      return cls(0, )
    if len(identifier) > 1:
      return cls._resolveNames(*identifier)
    identifier = identifier[0]
    if isinstance(identifier, (tuple, list, frozenset, set)):
      return cls._resolveNames(*identifier)
    if isinstance(identifier, int):
      return cls._resolveIndex(identifier)
    if isinstance(identifier, str):
      return cls._resolveName(identifier)
    return cls._resolveValue(identifier)



"""EZHook collects the field entries in EZData class bodies. """
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..mcls import AbstractNamespace as ASpace
from ..mcls.space_hooks import AbstractSpaceHook, ReservedNames
from ..utilities import textFmt, maybe, stringList
from ..waitaminute import TypeException, attributeErrorFactory
from ..waitaminute.ez import UnorderedEZException, EZDeleteException
from ..waitaminute.ez import UnfrozenHashException, FrozenEZException
from ..waitaminute.meta import ReservedName
from . import EZSlot

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Iterator, Callable, TypeAlias, Any, Never, Self

  from worktoy.ezdata import EZData

  Slots: TypeAlias = tuple[str, ...]
  SlotTypes: TypeAlias = Callable[[EZData, str], type]
  SlotDefaults: TypeAlias = Callable[[EZData, str], Any]

  Dunder: TypeAlias = Callable[[Any], Any]
  Factory: TypeAlias = Callable[[], Dunder]
  Factories: TypeAlias = dict[str, Factory]
  __INIT__: TypeAlias = Callable[[EZData], None]
  __LT__: TypeAlias = Callable[[EZData, EZData], bool]
  __LE__: TypeAlias = Callable[[EZData, EZData], bool]
  __GT__: TypeAlias = Callable[[EZData, EZData], bool]
  __GE__: TypeAlias = Callable[[EZData, EZData], bool]
  __EQ__: TypeAlias = Callable[[EZData, EZData], bool]
  __HASH__: TypeAlias = Callable[[EZData], int]
  __STR__: TypeAlias = Callable[[EZData], str]
  __REPR__: TypeAlias = Callable[[EZData], str]
  __ITER__: TypeAlias = Callable[[EZData], Iterator]
  __LEN__: TypeAlias = Callable[[EZData], int]
  __GETITEM__: TypeAlias = Callable[[EZData, str], Any]
  __SETITEM__: TypeAlias = Callable[[EZData, str, Any], None]
  __DELITEM__: TypeAlias = Callable[[EZData, str], Never]
  __GETATTR__: TypeAlias = Callable[[EZData, str], Any]
  __SETATTR__: TypeAlias = Callable[[EZData, str, Any], None]
  __DELATTR__: TypeAlias = Callable[[EZData, str], Never]
  AS_TUPLE: TypeAlias = Callable[[EZData], tuple[Any, ...]]
  AS_DICT: TypeAlias = Callable[[EZData], dict[str, Any]]


class EZSpaceHook(AbstractSpaceHook):
  """EZHook collects the field entries in EZData class bodies. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  __bad_names__ = """__slots__, __init__, __eq__, 
    __iter__, __getitem__, __setitem__, __getattr__, __len__, __hash__"""
  __auto_names__ = """__slots__, __init__, __eq__, __str__, __repr__,
    __iter__, __getitem__, __setitem__, __getattr__, __len__, __hash__"""

  #  Private Variables
  __new_callables__ = None
  __added_slots__ = None
  __typehint_mode__ = None
  __normal_mode__ = None

  #  Public Variables
  reservedNames = ReservedNames()
  if TYPE_CHECKING:  # pragma: no cover
    from . import EZSpace
    space: EZSpace

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getAutoNameFactoryDict(self, ) -> Factories:
    """Returns a dictionary of auto-named methods."""
    return {
      '__init__': self.initFactory,
      '__eq__': self.eqFactory,
      '__hash__': self.hashFactory,
      '__str__': self.strFactory,
      '__repr__': self.reprFactory,
      '__iter__': self.iterFactory,
      '__len__': self.lenFactory,
      '__getitem__': self.getItemFactory,
      '__setitem__': self.setItemFactory,
      '__delitem__': self.delItemFactory,
      '__getattr__': self.getAttrFactory,
      '__setattr__': self.setAttrFactory,
      '__delattr__': self.delAttrFactory,
      '__lt__': self.ltFactory,
      '__le__': self.leFactory,
      '__gt__': self.gtFactory,
      '__ge__': self.geFactory,
      'asTuple': self.asTupleFactory,
      'asDict': self.asDictFactory,
    }

  def _getBadNames(self) -> list[str]:
    """Returns a tuple of names that are reserved and should not be used."""
    return stringList(self.__bad_names__, )

  def _getAddedSlots(self) -> list[EZSlot]:
    """Returns a list of slots added by the current class body."""
    addedSlots = maybe(self.__added_slots__, [])
    ownerName = self.space.getClassName()
    out = []
    for ezSlot in addedSlots:
      if ezSlot.__owner_name__ == ownerName:
        out.append(ezSlot)
    return out

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def addSlot(self, ezSlot: EZSlot, **kwargs) -> None:
    """Adds a slot entry"""
    existing = self._getAddedSlots()
    if ezSlot in existing:
      return
    ezSlot.__owner_name__ = self.space.getClassName()
    self.__added_slots__ = [*existing, ezSlot]

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def preCompilePhase(self, compiledSpace: dict) -> dict:
    """Sets the factory created functions"""
    for name, factory in self._getAutoNameFactoryDict().items():
      compiledSpace[name] = factory()
    return compiledSpace

  def postCompilePhase(self, compiledSpace: dict) -> dict:
    """The postCompileHook method is called after the class is compiled."""
    ezSlots = self._getAddedSlots()
    if self.space.getKwargs().get('order', False) and ezSlots:
      for ezSlot in ezSlots:
        val = ezSlot.__default_value__
        try:
          _ = val < val or val > val
        except Exception as exception:
          if 'not supported between instances of ' in str(exception):
            clsName = self.space.getClassName()
            fName = ezSlot.name
            fType = ezSlot.typeValue
            raise UnorderedEZException(clsName, fName, fType)
          raise exception
    compiledSpace['__slot_objects__'] = ezSlots
    compiledSpace['__slots__'] = [ez.name for ez in ezSlots]
    return compiledSpace

  def setItemPhase(self, key: str, value: Any, oldValue: Any, ) -> bool:
    """
    Creates a slot entry for the given key and value, when value is not a
    callable, a descriptor or a special reserved name.
    """
    if key in self._getBadNames():
      if not hasattr(value, '__is_root__'):
        raise ReservedName(key)
      return True  # Ignores @_root decorated methods
    if key in self.reservedNames:
      return False  # Already handled by ReservedNameHook
    if callable(value):
      return False
    if hasattr(value, '__get__'):
      return False
    ezSlot = EZSlot(key)
    ezSlot.__type_value__ = type(value)
    ezSlot.__default_value__ = value
    self.addSlot(ezSlot)
    return True

  def preparePhase(self, space: ASpace, ) -> None:
    for base in self.space.getBases():
      for ezSlot in getattr(base, '__slot_objects__', []):
        self.addSlot(ezSlot)

  # \_____________________________________________________________________/ #
  #  Method factories
  # /¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨\ #
  @staticmethod
  def initFactory() -> __INIT__:
    """
    Creates the '__init__' method for the EZData class.
    """

    def __init__(self, *args, **kwargs) -> None:
      """
      The generated '__init__' method sets attributes on the instance
      based on given arguments. Keyword arguments take precedence.
      Positional arguments are applied in order.
      """
      posArgs = [*args, ]
      while len(posArgs) < len(self.__slot_objects__):
        posArgs.append(None)
      for (arg, slot) in zip(args, self.__slot_objects__):
        if arg is None:
          continue
        if not isinstance(arg, slot.typeValue):
          if slot.typeValue is not str:
            try:
              setattr(self, slot.name, slot.typeValue(arg))
            except Exception as e:
              raise TypeException('arg', arg, slot.typeValue, ) from e
            else:
              continue
          raise TypeException('arg', arg, slot.typeValue, )
        setattr(self, slot.name, arg)
      for key, val in kwargs.items():
        if key in self.__slots__:
          slot = self.__slot_objects__[self.__slots__.index(key)]
          if not isinstance(val, slot.typeValue):
            if slot.typeValue is not str:
              setattr(self, key, slot.typeValue(val))
              continue
            raise TypeException('val', val, slot.typeValue, )
          setattr(self, key, val)

    setattr(__init__, '__auto_generated__', True)
    return __init__

  @classmethod
  def ltFactory(cls, ) -> __LT__:
    """Creates the '__lt__' method for the EZData class. """

    def __lt__(self, other: Self) -> bool:
      if not type(self).isOrdered:
        raise UnorderedEZException(type(self).__name__)
      if type(self) is not type(other):
        return NotImplemented
      for slot in getattr(self, '__slots__'):
        selfVal, otherVal = getattr(self, slot), getattr(other, slot)
        if selfVal < otherVal:
          return True
        elif selfVal > otherVal:
          return False
        continue
      return False

    return __lt__

  @classmethod
  def leFactory(cls, ) -> __LE__:
    """Creates the '__le__' method for the EZData class. """

    def __le__(self, other: Self) -> bool:  # NOQA
      if not type(self).isOrdered:
        raise UnorderedEZException(type(self).__name__)
      if type(self) is not type(other):
        return NotImplemented
      for slot in getattr(self, '__slots__'):
        selfVal, otherVal = getattr(self, slot), getattr(other, slot)
        if selfVal < otherVal:
          return True
        elif selfVal > otherVal:
          return False
        continue
      return True

    return __le__

  @classmethod
  def gtFactory(cls, ) -> __GT__:
    """Creates the '__gt__' method for the EZData class."""

    def __gt__(self, other: Self) -> bool:
      if not type(self).isOrdered:
        raise UnorderedEZException(type(self).__name__)
      if type(self) is not type(other):
        return NotImplemented
      for slot in getattr(self, '__slots__'):
        selfVal, otherVal = getattr(self, slot), getattr(other, slot)
        if selfVal > otherVal:
          return True
        elif selfVal < otherVal:
          return False
        continue
      return False

    return __gt__

  @classmethod
  def geFactory(cls, ) -> __GE__:
    """Creates the '__ge__' method for the EZData class. """

    def __ge__(self, other: Self) -> bool:
      if not type(self).isOrdered:
        raise UnorderedEZException(type(self).__name__)
      if type(self) is not type(other):
        return NotImplemented
      for slot in getattr(self, '__slots__'):
        selfVal, otherVal = getattr(self, slot), getattr(other, slot)
        if selfVal > otherVal:
          return True
        elif selfVal < otherVal:
          return False
        continue
      return True

    return __ge__

  @staticmethod
  def eqFactory() -> __EQ__:
    """
    Creates the '__eq__' method for the EZData class.
    """

    def __eq__(self, other: EZData) -> bool:
      """
      Instances of EZData are equal if each of their data fields are equal.
      """
      if type(self) is not type(other):
        return NotImplemented
      for slot in self.__slot_objects__:
        if getattr(self, slot.name) != getattr(other, slot.name):
          return False
      return True

    setattr(__eq__, '__auto_generated__', True)
    return __eq__

  @staticmethod
  def hashFactory() -> __HASH__:
    """Creates the '__hash__' method for the EZData class."""

    def __hash__(self) -> int:
      """The hash of an EZData instance is the hash of its data fields."""
      if type(self).isFrozen:
        values = []
        for slot in self.__slots__:
          values.append(getattr(self, slot))
        return hash((*values,))
      clsName = self.__slot_objects__[0].ownerName
      raise UnfrozenHashException(clsName)

    setattr(__hash__, '__auto_generated__', True)
    return __hash__

  @staticmethod
  def strFactory() -> __STR__:
    """The strFactory method is called when the class is created."""

    def __str__(self) -> str:
      """The __str__ method is called when the class is created."""
      clsName = type(self).__name__
      names = [ezSlot.name for ezSlot in self.__slot_objects__]
      vals = [str(getattr(self, name)) for name in names]
      keyVals = ['%s=%s' % (name, val) for name, val in zip(names, vals)]
      return """%s(%s)""" % (clsName, ', '.join(keyVals))

    setattr(__str__, '__auto_generated__', True)
    return __str__

  @staticmethod
  def reprFactory(*ezSlots) -> __REPR__:
    """The reprFactory method is called when the class is created."""

    def __repr__(self) -> str:
      """The __repr__ method is called when the class is created."""
      infoSpec = """%s(%s)"""
      clsName = type(self).__name__
      fieldNames = [ezSlot.name for ezSlot in self.__slot_objects__]
      fieldValues = [getattr(self, name) for name in fieldNames]
      fieldRepr = []
      for field in fieldValues:
        if isinstance(field, str):
          fieldRepr.append("""'%s'""" % field)
          continue
        fieldRepr.append(str(field))
      if fieldRepr:
        info = infoSpec % (clsName, ', '.join(fieldRepr))
      else:
        info = infoSpec % (clsName, '')
      return textFmt(info)

    setattr(__repr__, '__auto_generated__', True)
    return __repr__

  @staticmethod
  def iterFactory() -> __ITER__:
    """The iterFactory method is called when the class is created."""

    def __iter__(self, ) -> Iterator:
      """Implementation of the iteration protocol"""
      for key in self.__slots__:
        yield getattr(self, key)

    setattr(__iter__, '__auto_generated__', True)
    return __iter__

  @staticmethod
  def lenFactory() -> __LEN__:
    """The lenFactory method is called when the class is created."""

    def __len__(self) -> int:
      """The __len__ method is called when the class is created."""
      return len(self.__slots__)

    setattr(__len__, '__auto_generated__', True)
    return __len__

  @staticmethod
  def getItemFactory() -> __GETITEM__:
    """The getItemFactory method is called when the class is created."""

    def __getitem__(self, identifier: str) -> Any:
      """The __getitem__ method is called when the class is created."""
      if isinstance(identifier, int):
        if identifier < 0:
          return self[identifier + len(self)]
        if identifier < len(self):
          #  Uses the 'int' valued identifier to retrieve the name from
          #  the '__slots__' tuple. Then this name is passed recursively
          #  back to '__getitem__' retrieving the value from 'str'.
          return self[self.__slots__[identifier]]
        infoSpec = """Index %d out of range for '%s' with %d slots."""
        clsName = type(self).__name__
        info = infoSpec % (identifier, clsName, len(self.__slots__))
        raise IndexError(textFmt(info))
      if isinstance(identifier, str):
        #  Uses the 'str' valued identifier to retrieve the value. The
        #  identifier may initially have been an 'int' value, resolved to
        #  a 'str' value as described above.
        if identifier in self.__slots__:
          return getattr(self, identifier)
        e = attributeErrorFactory(type(self), identifier)
        raise KeyError(identifier) from e
      if isinstance(identifier, slice):
        sliceKeys = self.__slots__[identifier]
        out = []
        for sliceKey in sliceKeys:
          out.append(self[sliceKey])
        return (*out,)
      raise TypeException('key', identifier, str, int, slice)

    setattr(__getitem__, '__auto_generated__', True)
    return __getitem__

  @staticmethod
  def setItemFactory() -> __SETITEM__:

    def __setitem__(self, identifier: Any, value: Any) -> None:
      """
      Assigns value(s) to this object’s slots via integer index, slot name,
      or slice.

      - If 'identifier' is an int, it is interpreted as a positional index
        into the slot sequence. Negative indices are supported. Raises
        IndexError if out of range.

      - If 'identifier' is a str, it must match one of the defined slot
        names, otherwise KeyError is raised.

      - If 'identifier' is a slice, it refers to a contiguous subset of slot
        names. The 'value' assigned must be an iterable of equal length. If
        lengths differ, raises IndexError. If a string is provided as the
        'value', raises TypeError with an explanatory message—this
        prevents the common error of assigning an iterable of characters,
        but bytes and bytearray are allowed.

      Raises TypeException if 'identifier' is not an int, str, or slice.
      """

      if isinstance(identifier, int):
        if identifier < 0:
          return self.__setitem__(identifier + len(self), value)
        if identifier < len(self):
          return self.__setitem__(self.__slots__[identifier], value)
        infoSpec = """Index %d out of range for '%s' with %d slots."""
        clsName = type(self).__name__
        info = infoSpec % (identifier, clsName, len(self))
        raise IndexError(textFmt(info))
      if isinstance(identifier, str):
        if identifier in self.__slots__:
          return setattr(self, identifier, value)
        raise KeyError(identifier)
      if isinstance(identifier, slice):
        if isinstance(value, str):
          infoSpec = """Tried setting slice: 'self[%s]' to a 'str' object: 
          '%s'. Because this is nearly always not meant to be taken as an 
          iterable of characters, 'EZData' does not allow this despite the 
          fact that Python technically does. Please note that this 
          prohibition applies only to 'str' not to 'bytes' or 'bytearray'."""
          info = textFmt(infoSpec % (identifier, value))
          raise TypeError(info)
        sliceSlots = self.__slots__[identifier]
        if len(sliceSlots) != len(value):
          infoSpec = """Slice '%s' of '%s' with %d slots cannot be set to 
          value of length %d!"""
          clsName = type(self).__name__
          lenSlots = len(self)
          lenValue = len(value)
          info = infoSpec % (identifier, clsName, lenSlots, lenValue)
          raise IndexError(textFmt(info))
        for (slot, val) in zip(sliceSlots, value):
          self.__setitem__(slot, val)
        else:
          return
      raise TypeException('key', identifier, str, int, slice)

    setattr(__setitem__, '__auto_generated__', True)
    return __setitem__

  @staticmethod
  def delItemFactory() -> __DELITEM__:
    """Creates '__delitem__' method that always raise TypeError"""

    def __delitem__(self, key: Any) -> Never:
      infoSpec = """EZData classes does not support deletion of slots, 
      but received "del self['%s']" on class: '%s'!"""
      clsName = type(self).__name__
      info = infoSpec % (key, clsName)
      raise TypeError(textFmt(info))

    setattr(__delitem__, '__auto_generated__', True)
    return __delitem__

  @staticmethod
  def getAttrFactory() -> __GETATTR__:
    """The getAttrFactory method is called when the class is created."""

    def __getattr__(self, key: str) -> Any:
      """If the given key is in one of the '__slots__', but this method
      still is invoked, it means that 'object.__getattribute__' was unable
      to retrieve a value for the given key. This method confirms that no
      value is present at 'self' nor in any base. In this case, the method
      attempts to set a default value at the given key for 'self'. """

      if key not in self.__slots__:  # Raises
        raise attributeErrorFactory(self, key)
      value = None
      exception = None
      slot = [s for s in self.__slot_objects__ if s.name == key][0]
      value = getattr(slot, 'defaultValue')
      setattr(self, key, value)
      return value

    setattr(__getattr__, '__auto_generated__', True)
    return __getattr__

  @staticmethod
  def setAttrFactory() -> __SETATTR__:
    """The setAttrFactory method is called when the class is created."""

    def __setattr__(self, key: str, value: Any) -> None:
      """Sets the value of the given key in the EZData instance."""
      if key not in self.__slots__:
        raise attributeErrorFactory(self, key)
      if type(self).isFrozen:
        try:
          oldValue = object.__getattribute__(self, key, )
        except AttributeError:
          return object.__setattr__(self, key, value)
        else:
          slot0 = self.__slot_objects__[0]
          clsName = slot0.ownerName
          oldValue = getattr(self, key, )
          raise FrozenEZException(key, clsName, oldValue, value)
      return object.__setattr__(self, key, value)

    setattr(__setattr__, '__auto_generated__', True)
    return __setattr__

  @staticmethod
  def delAttrFactory() -> __DELATTR__:
    """The delAttrFactory method is called when the class is created."""

    def __delattr__(self, key: str) -> Never:
      """Illegal deleter"""
      raise EZDeleteException(type(self), key, )

    setattr(__delattr__, '__auto_generated__', True)
    return __delattr__

  @staticmethod
  def asTupleFactory() -> AS_TUPLE:
    """The asTupleFactory method is called when the class is created."""

    def asTuple(self, ) -> tuple:
      """Returns a tuple of the values of the EZData instance."""
      return tuple(getattr(self, key) for key in self.__slots__)

    setattr(asTuple, '__auto_generated__', True)
    return asTuple

  @staticmethod
  def asDictFactory() -> AS_DICT:
    """The asDictFactory method is called when the class is created."""

    def asDict(self: EZData) -> dict:
      """Returns a dictionary of the values of the EZData instance."""
      return {key: getattr(self, key) for key in self.__slots__}

    setattr(asDict, '__auto_generated__', True)
    return asDict



"""
KeeFlagsSpace subclasses KeeSpace from the worktoy.keenum package
providing the namespace object required for KeeFlags.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..mcls import BaseSpace, AbstractNamespace
from ..utilities import maybe
from ..waitaminute.keenum import KeeFlagDuplicate
from . import KeeFlag, KeeFlagsHook

if TYPE_CHECKING:  # pragma: no cover
  from typing import Self, Type, TypeAlias, Callable, Self

  from . import KeeFlags
  from . import KeeFlagsMeta

  KFMType: TypeAlias = Type[KeeFlagsMeta]
  Bases: TypeAlias = tuple[Type, ...]
  GetKeeFlags: TypeAlias = Callable[[Self], dict[str, KeeFlag]]


class KeeFlagsSpace(BaseSpace):
  """
  KeeFlagsSpace subclasses KeeSpace from the worktoy.keenum package
  providing the namespace object required for KeeFlags.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __kee_flags__ = None
  __base_flags__ = None

  #  Space Hooks
  keeFlagsHook = KeeFlagsHook()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getBaseFlags(self, ) -> dict[str, KeeFlag]:
    return maybe(self.__base_flags__, dict())

  def getKeeFlags(self, ) -> dict[str, KeeFlag]:
    baseFlags = self._getBaseFlags()
    keeFlags = maybe(self.__kee_flags__, dict())
    for name, keeFlag in keeFlags.items():
      baseFlags[name] = keeFlag
    return baseFlags

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def addBaseFlag(self, name: str, keeFlag: KeeFlag, **kwargs) -> None:
    """Adds a """
    existing = self._getBaseFlags()
    existing[name] = keeFlag
    self.__base_flags__ = existing

  def addKeeFlag(self, name: str, keeFlag: KeeFlag, **kwargs) -> None:
    baseFlags = self._getBaseFlags()
    keeFlags = self.getKeeFlags()
    if name in maybe(self.__kee_flags__, dict()):
      oldFlag = self.__kee_flags__[name]
      raise KeeFlagDuplicate(name, oldFlag, keeFlag)
    keeFlag.__member_index__ = len(baseFlags) + len(keeFlags)
    keeFlag.__member_name__ = name
    keeFlag.__field_name__ = name
    keeFlags[name] = keeFlag
    self.__kee_flags__ = keeFlags

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, mcls: KFMType, name: str, bases: Bases, **kw) -> None:
    if name == 'KeeFlags':
      BaseSpace.__init__(self, mcls, name, bases, **kw)
    else:
      AbstractNamespace.__init__(
        self,
        mcls,
        name,
        bases,
        _strictMRO=False,
        **kw
      )
      self.__kee_flags__ = None
      cls = type(self)
      for base in bases:
        baseSpace = getattr(base, '__namespace__', None)
        for name, keeFlag in baseSpace.getKeeFlags().items():
          self.addBaseFlag(name, keeFlag)
      for space in self.getMRONamespaces():
        for name, sigFunc in getattr(space, '__overload_map__', ).items():
          self.__overload_map__[name] = dict()
          for sig, func in sigFunc.items():
            self.__overload_map__[name][sig] = func
        for name, func in getattr(space, '__fallback_map__', ).items():
          existing = maybe(self.__fallback_map__, dict())
          existing[name] = func
          self.__fallback_map__ = existing
        for name, func in getattr(space, '__finalizer_map__', ).items():
          existing = maybe(self.__finalizer_map__, dict())
          existing[name] = func
          self.__finalizer_map__ = existing

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def _getKeeFlagsFactory(cls, ) -> GetKeeFlags:
    """
    Factory function creating the getter function for the dictionary of
    the 'KeeFlag' objects forming the basis of the 'KeeFlags' class.
    """

    def func(cls_: Type[KeeFlags]) -> dict[str, KeeFlag]:
      out = dict()
      for i, (name, flag) in enumerate(cls_.__kee_flags__.items()):
        out[name] = flag.clone(cls_, i)
      return out

    setattr(func, '__name__', 'getKeeFlags')
    docSpec = """Getter function for the flags dictionary."""
    setattr(func, '__doc__', docSpec)
    return func

  def postCompile(self, namespace: dict) -> dict:
    """
    Post compile is called after the class has been created.
    This method is used to finalize the namespace.
    """
    namespace = BaseSpace.postCompile(self, namespace)
    if self.getClassName() == 'KeeFlags':
      return namespace
    namespace['__kee_flags__'] = self.getKeeFlags()
    namespace['__kee_members__'] = None
    namespace['getKeeFlags'] = classmethod(self._getKeeFlagsFactory())
    return namespace



"""
EZDeleteException provides a custom exception class raised to indicate an
attempt to delete a field in an EZData subclass. EZData subclasses do not
allow deletion of fields.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class EZDeleteException(TypeError):
  """
  EZDeleteException provides a custom exception class raised to indicate an
  attempt to delete a field in an EZData subclass. EZData subclasses do not
  allow deletion of fields.
  """

  __slots__ = ('name', 'cls')

  def __init__(self, cls: type, name: str, ) -> None:
    self.cls = cls
    self.name = name
    TypeError.__init__(self, )

  def __str__(self) -> str:
    infoSpec = """Attempted to delete field '%s' in EZData subclass '%s'. 
    EZData subclasses do not allow deletion of fields. """
    fName = self.name
    clsName = self.cls.__name__
    info = infoSpec % (fName, clsName)
    return textFmt(info, )

  __repr__ = __str__



"""DispatchException provides a custom exception raised when an instance
of OverloadDispatcher fails to resolve the correct function from the
given arguments. Because the overload protocol relies on type matching,
this exception subclasses TypeError such that it can be caught by external
error handlers. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, TypeAlias

  Args: TypeAlias = tuple[Any, ...]
  Excs: TypeAlias = tuple[Exception, ...]

  from worktoy.dispatch import Dispatcher


class DispatchException(TypeError):
  """DispatchException provides a custom exception raised when an instance
  of OverloadDispatcher fails to resolve the correct function from the
  given arguments. Because the overload protocol relies on type matching,
  this exception subclasses TypeError such that it can be caught by external
  error handlers. """

  __slots__ = ('dispatch', 'args', 'excs')

  def __init__(self, dispatch: Dispatcher, args: Args, ) -> None:
    self.dispatch = dispatch
    self.args = args
    TypeError.__init__(self, )

  def __str__(self) -> str:
    """
    Return a string representation of the DispatchException.
    """
    infoSpec = """Dispatcher object: '%s' failed to dispatch arguments: 
    <br><tab>%s<br><tab>matching type signature: '%s'<br>"""
    dispStr = str(self.dispatch)
    argsStr = ', '.join(str(arg) for arg in self.args)
    from ...dispatch import TypeSig
    typeStr = str(TypeSig.fromArgs(*self.args))
    info = infoSpec % (dispStr, argsStr, typeStr)
    return textFmt(info, )

  __repr__ = __str__



"""The 'joinWords' function joins a list of words into a single string
with appropriate use of commas and 'and/or'. By default, the final two
given words a separated by 'and', but this can be changed at keyword
argument 'sep'."""
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations


def joinWords(*words: str, **kwargs) -> str:
  """Join a list of words into a single string with appropriate use of
  commas and 'and'."""
  if not words:
    return ''
  sep = kwargs.get('sep', 'and')
  if len(words) == 1:
    if isinstance(words[0], (list, tuple)):
      return joinWords(*words[0])
    return str(words[0])
  if len(words) == 2:
    return '%s %s %s' % (words[0], sep, words[1])
  return joinWords(', '.join(words[:-1]), words[-1])



"""
KeeIndexError provides a custom exception raised to indicate that a given
positive index is out of range for a particular enumeration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class KeeIndexError(IndexError):
  """
  KeeIndexError provides a custom exception raised to indicate that a given
  positive index is out of range for a particular enumeration.
  """

  __slots__ = ('keenum', 'index')

  def __init__(self, keenum: Any, index: int) -> None:
    self.keenum, self.index = keenum, index

  def __str__(self) -> str:
    infoSpec = """KeeNum object: '%s' does not have an index '%d'!"""
    info = infoSpec % (self.keenum, self.index)
    from ...utilities import textFmt
    return textFmt(info)

  __repr__ = __str__



"""
KeeCaseException provides a custom exception raised to indicate that a
member of an enumeration was set not with upper case.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  pass


class KeeCaseException(ValueError):
  """
  KeeCaseException provides a custom exception raised to indicate that a
  member of an enumeration was set not with upper case.
  """

  __slots__ = ('name',)

  def __init__(self, name: str, ) -> None:
    """Initialize the KeeCaseException object."""
    self.name = name
    ValueError.__init__(self, )

  def __str__(self) -> str:
    """Return the string representation of the KeeCaseException object."""
    infoSpec = """KeeNum members must have upper case names, but received: 
    '%s'"""
    from ...utilities import textFmt
    return textFmt(infoSpec % self.name)

  __repr__ = __str__



"""
KeeMember encapsulates a member of an enumeration.

Defining properties: (independent properties)
  - name: The name of the member. This is the name by which the member
  appears in the class body of the enumeration. It is the name passed to
  the '__set_name__' method. These must be unique with enumerations and
  must be uppercase. The case requirement is more than convention,
  it is enforced. When an enumeration class body contains key, value pairs
  with the key not in uppercase, it is understood to mean that the value
  does not represent a member of the enumeration.
  - value: The value of the member. Uniquely, this value is not required
  to be unique across members of the same enumeration.
  - index: The index of the member. This index specifies how many members
  are before this member in the enumeration.

KeeMember may be used directly or may be further subclassed. The KeeNum
classes are created by the KeeMeta class which define the class behaviour
post creation, and KeeMember define how members are included in the
enumeration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..desc import Field, AttriBox
from ..utilities import textFmt
from ..waitaminute import VariableNotNone, MissingVariable, TypeException
from ..waitaminute.keenum import KeeCaseException

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Type, TypeAlias

  KEENUM: TypeAlias = Type[object]


class Kee(AttriBox):
  """KeeMember encapsulates a member of an enumeration. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables
  __context_instance__ = None
  __context_owner__ = None

  #  Fallback Variables

  #  Private Variables
  __field_value__ = None
  __num_index__ = None
  __num_name__ = None

  #  Public Variables
  name = Field()
  index = Field()
  value = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @value.GET
  def getValue(self, **kwargs) -> Any:
    """
    Uses the 'AttriBox' lazy instantiation mechanism.
    """
    if self.__field_value__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self.__field_value__ = self._createFieldObject()
      return self.getValue(_recursion=True)
    fieldType = self.getFieldType()
    if isinstance(self.__field_value__, fieldType):
      return self.__field_value__
    raise TypeException('__field_value__', self.__field_value__, fieldType)

  @name.GET
  def _getName(self) -> str:
    """Set by 'Object.__set_name__' when the enumeration is created."""
    if self.__num_name__ is None:
      raise MissingVariable(self, '__num_name__', str)
    if isinstance(self.__num_name__, str):
      return self.__num_name__
    raise TypeException('__num_name__', self.__num_name__, str)

  @index.GET
  def _getIndex(self) -> int:
    """Set by 'KeeSpace.addNum' when the enumeration is created."""
    if self.__num_index__ is None:
      raise MissingVariable(self, '__num_index__', int)
    if isinstance(self.__num_index__, int):
      return self.__num_index__
    raise TypeException('__num_index__', self.__num_index__, int)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  The 'KeeSpace' namespace class expects to be able to set 'name' and
  #  'index'  before the enumeration class is created. Subclasses that
  #  change this behaviour must also implement these changes  in the
  #  'KeeSpace' class.

  @name.SET
  def _setName(self, name: str) -> None:
    if not name.isupper():
      raise KeeCaseException(name)
    if self.__num_name__ is not None:
      raise VariableNotNone('name', self.__num_name__)
    self.__num_name__ = name

  @index.SET
  def _setIndex(self, index: int) -> None:
    if not isinstance(index, int):
      raise TypeException('index', index, int)
    if self.__num_index__ is not None:
      raise VariableNotNone('index', self.__num_index__)
    self.__num_index__ = index

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __int__(self, ) -> int:
    return self.index

  def __str__(self) -> str:
    """Return the string representation of the member."""
    infoSpec = """<%s member: %s>"""
    try:
      keeName = self.name
    except AttributeError:
      keeName = '- N/A -'
    clsName = type(self).__name__
    info = infoSpec % (clsName, keeName)
    return textFmt(info)

  __repr__ = __str__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Parent Methods   # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getContextInstance(self) -> Any:
    return None



"""
KeeSpaceHook collects the KeeMember objects encountered in the class
bodies of KeeNum classes. This happens during the 'setItemPhase'. To avoid
context leakage, the members are collected in the owning namespace object.
The namespace object is expected to implement a method called
'addEnumeration' which 'KeeHook' calls to register the members. The
'KeeHook' provides no further functionality than deciding what key,
value pairs to collect as future members of the enumeration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..mcls.space_hooks import AbstractSpaceHook
from . import Kee

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Type, TypeAlias, Callable
  from . import KeeMeta

  __INIT__: TypeAlias = Callable[[KeeMeta, Kee], None]
  __GET__: TypeAlias = Callable[[KeeMeta, KeeMeta, Type[KeeMeta]], Any]


class KeeSpaceHook(AbstractSpaceHook):
  """
  KeeSpaceHook collects the KeeMember objects encountered in the class
  bodies of KeeNum classes. This happens during the 'setItemPhase'. To avoid
  context leakage, the members are collected in the owning namespace object.
  The namespace object is expected to implement a method called
  'addEnumeration' which 'KeeHook' calls to register the members. The
  'KeeSpaceHook' provides no further functionality than deciding what key,
  value pairs to collect as future members of the enumeration.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def setItemPhase(self, key: str, val: Any, old: Any = None, ) -> bool:
    """
    Routes 'Kee' objects to the 'addNum' method of the present namespace
    object.
    """
    if isinstance(val, Kee):
      self.space.addNum(key, val)
      return True
    return False



"""
KeeNameError provides a custom exception raised to indicate an attempt to
resolve a member of an enumeration with an unrecognized name.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class KeeNameError(KeyError):
  """
  KeeNameError provides a custom exception raised to indicate an attempt to
  resolve a member of an enumeration with an unrecognized name.
  """

  __slots__ = ('keenum', 'name')

  def __init__(self, keenum: Any, name: str) -> None:
    self.keenum, self.name = keenum, name

  def __str__(self) -> str:
    infoSpec = """KeeNum enumeration: '%s' has no member named '%s'!"""
    info = infoSpec % (self.keenum, self.name)
    from ...utilities import textFmt
    return textFmt(info)

  __repr__ = __str__



"""The 'worktoy' package provides a collection of utilities leveraging
advanced python features including custom metaclasses and the descriptor
protocol. The readme file included provides detailed documentation on the
included features. The modules provided depend on each other in
implementation, but can be used independently.

The package consists of thr following modules:
- utilities: A set of general-purpose utility functions and classes.
- waitaminute: Tools for managing execution flow and timing.
- core: Core functionalities and base classes for the package.
- desc: Descriptor protocol utilities.
- dispatch: Function and method dispatching used by overload system.
- mcls: Custom metaclass implementations.
- keenum: Enumeration utilities.
- ezdata: Dataclass implementation.
- work_io: Input/output utilities for file and directory management.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from . import utilities
from . import waitaminute
from . import core
from . import desc
from . import dispatch
from . import mcls
from . import keenum
from . import ezdata
from . import work_io

__all__ = [
  'utilities',
  'waitaminute',
  'core',
  'desc',
  'dispatch',
  'mcls',
  'keenum',
  'ezdata',
  'work_io',
]



"""EZSlot encapsulates data fields in the EZData class."""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..desc import Field
from ..mcls import BaseObject
from ..utilities import maybe

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Self


class EZSlot(BaseObject):
  """EZSlot encapsulates data fields in the EZData class. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __future_name__ = None
  __type_value__ = None
  __default_value__ = None
  __global_scope__ = None
  __owner_name__ = None

  #  Public Variables
  ownerName = Field()
  name = Field()
  typeValue = Field()
  defaultValue = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @name.GET
  def _getName(self) -> str:
    """Get the name of the slot."""
    return self.__future_name__

  @typeValue.GET
  def _getTypeValue(self, **kwargs) -> type:
    """Get the type value of the slot."""
    return self.__type_value__

  @defaultValue.GET
  def _getDefaultValue(self, **kwargs) -> Any:
    """Get the default value of the slot."""
    return self.__default_value__

  @ownerName.GET
  def _getOwnerName(self) -> str:
    """Get the owner name of the slot."""
    return self.__owner_name__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, name: str) -> None:
    self.__future_name__ = name

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __eq__(self, other: Self) -> bool:
    """
    Compares only against other 'EZSlot' instances, but only by the 'name'
    attribute.
    """
    cls = type(self)
    if isinstance(other, cls):
      if self.name == other.name:
        return True
      return False
    return NotImplemented

  def __str__(self) -> str:
    """Get the string representation of the EZSlot."""
    infoSpec = """%s<%s.%s>(%s: %s)"""
    clsName = type(self).__name__
    valName = maybe(self.__default_value__, '[NONE]')
    typeName = self.typeValue.__name__
    info = infoSpec % (clsName, self.ownerName, self.name, typeName, valName)
    return info

  __repr__ = __str__



"""
FrozenEZException provides a custom exception class raised to indicate an
attempt to change the value of a field in an EZData subclass that is
frozen, meaning it is immutable.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class FrozenEZException(TypeError):
  """
  FrozenEZException provides a custom exception class raised to indicate an
  attempt to change the value of a field in an EZData subclass that is
  frozen, meaning it is immutable.
  """

  __slots__ = ('fieldName', 'className', 'oldValue', 'newValue')

  def __init__(self, *args) -> None:
    fName, clsName, old, new_, *_ = [*args, None, None, None]
    self.fieldName = fName
    self.className = clsName
    self.oldValue = old
    self.newValue = new_
    TypeError.__init__(self, )

  def __str__(self) -> str:
    infoSpec = """Attempted to set value of field '%s' in frozen EZData 
    subclass: '%s' from '%s' to '%s'. """
    fName = self.fieldName
    clsName = self.className
    oldStr = repr(self.oldValue)
    newStr = repr(self.newValue)
    info = infoSpec % (fName, clsName, oldStr, newStr)
    return textFmt(info, )

  __repr__ = __str__



"""
BaseObject is the standard entry point for using the worktoy library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import BaseMeta
from ..core import Object

if TYPE_CHECKING:  # pragma: no cover
  pass


class BaseObject(Object, metaclass=BaseMeta):
  """
  BaseObject combines the core functionality of Object with the
  hook-based metaclass behavior of BaseMeta.

  From Object, it inherits robust constructor handling, controlled
  descriptor mutation, and context-aware access to instance and owner
  information. From BaseMeta, it gains support for function overloading
  and other hook-based class construction features via BaseSpace.

  Subclass this when you want both: sane, safe object semantics and
  overload-aware metaclass support.
  """
  pass



"""
The 'replaceFlex' function provides a vastly superior alternative to the
'str.replace' function.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from . import maybe


def replaceFlex(text: str, old: str, new: str, n: int = None) -> str:
  """
  Replaces only the nth occurrence of 'old' with 'new' in text.
  n is 1-based.
  """
  n = maybe(n, 1)
  i = -1
  for _ in range(n):
    i = text.find(old, i + 1)
    if i == -1:
      return text
  return text[:i] + new + text[i + len(old):]



"""
The 'worktoy.waitaminute.ez' module provides custom exception classes used
by the 'worktoy.ezdata' module.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._ez_multiple_inheritance import EZMultipleInheritance
from ._unordered_ez_exception import UnorderedEZException
from ._unfrozen_hash_exception import UnfrozenHashException
from ._frozen_ez_exception import FrozenEZException
from ._ez_delete_exception import EZDeleteException

___all__ = [
  'EZMultipleInheritance',
  'UnorderedEZException',
  'UnfrozenHashException',
  'FrozenEZException',
  'EZDeleteException',
]



"""
The 'resolveMRO' function takes any number of classes and creates a tuple
of classes having the same method resolution order (MRO) as a potential
new class inheriting from all of the in the order passed. For example,

class Foo(Tom, Dick, Harry):
  pass

predictedMRO = resolveMRO(Tom, Dick, Harry)
actualMRO = Foo.mro()
predictedMRO == actualMRO
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import joinWords, textFmt

if TYPE_CHECKING:  # pragma no cover
  pass


def resolveMRO(*bases: type, **kwargs) -> list[type]:
  if not bases:
    return []
  if len(bases) == 1:
    return [*bases[0].__mro__, ]
  baseMROs = [base.__mro__ for base in bases]
  maxC = sum([sum([1 for _ in b], ) for b in baseMROs])
  out = []
  _c = kwargs.get('_start', 0)
  while baseMROs:
    for baseMRO in baseMROs:
      base = baseMRO[0]
      tailMROs = [b for b in baseMROs if len(b) > 1]
      tailMROs = [b[1:] for b in tailMROs]
      for tail in tailMROs:
        if base in tail:
          break
      else:
        break
    else:
      infoSpec = """The bases received: (%s) cannot form a consistent mro!"""
      baseNames = joinWords(*[b.__name__ for b in bases], )
      info = textFmt(infoSpec % baseNames)
      raise TypeError(info)
    maybeEmpty = [[b for b in bs if b is not base] for bs in baseMROs]
    baseMROs = [b for b in maybeEmpty if b]
    out.append(base)
    _c += 1
    if _c > maxC:
      break
  else:
    return out
  raise RecursionError



"""
KeeFlagsHook provides a namespace hook for the 'KeeFlagsSpace' class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..mcls.space_hooks import AbstractSpaceHook
from . import KeeFlag

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class KeeFlagsHook(AbstractSpaceHook):
  """
  KeeFlagsHook provides a namespace hook for the 'KeeFlagsSpace' class.

  It intercepts key, KeeFlag pairs and redirects them to the 'addKeeFlag'
  method on the 'KeeFlagsSpace' instance.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def setItemPhase(self, key: str, val: Any, old: Any = None, ) -> bool:
    """
    Hook for setItem. This is called before the __setitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False.
    """
    if isinstance(val, KeeFlag):
      self.space.addKeeFlag(key, val)
      return True
    return False



"""
Sentinels represent special situations. The sentinel itself provides no
information about its meaning or function. It is merely a stateless unique
object equal only to itself.

Contents:

  - Sentinel: Base class for all sentinels. It derives from a custom
  metaclass providing functionality preventing instantiation and
  duplication.
  - DELETED: Sentinel used to indicate that an element has been deleted.
  Used by custom descriptors to implement deletion semantics: to 'delete'
  a custom descriptor from an instance, the descriptor 'sets' the value
  for the instance to 'DELETED'. The same descriptor then raises the
  appropriate 'AttributeError' when '__get__' would return 'DELETED'.
  - THIS: Allows references to classes from within the class bodies. Used
  by the 'AttriBox' descriptors and the '@Overload' decorators to specify
  an instance of the class. Similar to 'typing.Self'.
  - OWNER: Similar to THIS, but specifying the class itself, rather than
  an instance of it.
  - DESC: Used by 'AttriBox' along with THIS and OWNER, specifying the
  present descriptor.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from ._sentinel import Sentinel
from ._place_holder import PLACE_HOLDER
from ._deleted import DELETED
from ._owner import OWNER
from ._this import THIS
from ._desc import DESC
from ._wild_card import WILDCARD
from ._meta_call import METACALL
from ._fallback import FALLBACK
from ._function import Function

__all__ = [
  'Sentinel',
  'PLACE_HOLDER',
  'DELETED',
  'OWNER',
  'THIS',
  'DESC',
  'WILDCARD',
  'METACALL',
  'FALLBACK',
  'Function',
]



"""
MissingVariable is a custom exception class raise to indicate that a
variable has not been assigned a value when accessed. Generally indicates
that the variable has no fallback value.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class MissingVariable(AttributeError):
  """
  MissingVariable is a custom exception class raise to indicate that a
  variable has not been assigned a value when accessed. Generally indicates
  that the variable has no fallback value.
  """

  __slots__ = ('instance', 'varName', 'type_')

  def __init__(self, instance: Any, name: str, type_: type) -> None:
    self.varName = name
    self.instance = instance
    self.type_ = type_
    AttributeError.__init__(self, )

  def __str__(self) -> str:
    infoSpec = """Missing variable '%s' of type '%s' in instance '%s'!"""
    instanceName = getattr(self.instance, '__name__', 'Unknown')
    info = infoSpec % (self.varName, self.type_.__name__, instanceName)
    return textFmt(info)

  __repr__ = __str__



"""
KeeFlags enumerations all combinations of boolean valued flags. It
dynamically adds instances of KeeFlags for each boolean valued entry. The
result is a KeeNum-like enumeration consisting of all possible combinations
of boolean flags.

Each enumeration of the KeeFlags class is created at class creation time.
This means that the number of enumerations increase exponentially with
number of flags included. This presents no problem for the intended uses.
For example the FileAccess enumeration used in the test suite:


class FileAccess(KeeFlags):
  #  FileAccess demonstrates real-world bitmask flags for file permissions.

  #  Enumerations
  READ = Kee[int](0b0001)
  WRITE = Kee[int](0b0010)
  EXECUTE = Kee[int](0b0100)
  DELETE = Kee[int](0b1000)

  __null_value__ = 0b1111

  - The NULL Member -
While KeeNum enumerations may implement a member called 'NULL', KeeFlags
enumerations automatically has a member called 'NULL' which equals the
bitwise negation of all flags defined in the enumeration. KeeFlags
enumerations may provide a specific value '__null_value__'. If this value
is passed to the member resolution workflow, and no member equals it,
the 'fromValue' method will return the NULL member. If the
'__null_value__' does equal the value of a member, that member is returned
instead.

  - Flags and Names -
Each KeeFlags enumeration has a unique combination of flags that are HIGH.
The 'flags' and 'names' descriptors when accessed through a member returns
a list of the flags or names respectively that are HIGH for that member.

For example, 'FileAccess.READ_WRITE.flags' returns the '.READ' and
'.WRITE' single bit flags that are HIGH for the 'FileAccess.READ_WRITE'
member. The 'names' descriptor returns the names of the flags that
are HIGH for the member, such as 'READ' and 'WRITE'.

  - Flexibility -
Each enumeration of flags is generated automatically. This raises a
question of naming. Is it 'FileAccess.READ_EXECUTE' or is it
'FileAccess.EXECUTE_READ'? It follows the order of appearance in the class
body. However, while

"""
#  AGPL-3.0 license
#  Copyright (c) 2025-2026 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..desc import Field
from ..utilities import textFmt
from ..waitaminute import MissingVariable, TypeException
from . import KeeFlag, KeeFlagsMeta

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Iterator, Self


class KeeFlags(metaclass=KeeFlagsMeta):
  """
  KeeFlags is a metaclass that dynamically creates instances of KeeFlags
  for each boolean valued entry. It allows for the creation of an
  enumeration consisting of all possible combinations of boolean flags.

  Important Attributes:
  - flags: A descriptor returning the flags of a particular enumeration
  that are HIGH.

  Entries must be integer valued.

  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Class Variables (type hints)
  __member_list__: list[Self]
  __member_dict__: dict[frozenset[str], Self]

  #  Private Variables
  __member_index__ = None
  __member_value__ = None
  __frozen_state__ = None

  #  Public Variables
  index = Field()

  #  Virtual Variables
  lows = Field()
  highs = Field()
  value = Field()
  name = Field()
  names = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @value.GET
  def _getValue(self) -> Any:
    """
    Returns the object at the 'value' attribute of the member. The base
    implementation provides for retrieving a member from the class from a
    given value. Other than this, the attribute provides no further
    functionality. Subclasses may override this method to provide any
    object from any member. While not enforced, it is recommended that
    that member values should be of the same type and that the type should
    be immutable. Subclasses are free to grant multiple members the same
    value. In this case, the 'fromValue' method returns the first member
    having the given value.

    By default, the value is the index of the member.
    """
    return self.index

  @lows.GET
  def _getLows(self) -> Iterator[KeeFlag]:
    if self.__member_index__:
      for flag in type(self).flags:
        if self.index & (1 << flag.index):
          continue
        yield flag
    else:
      yield from type(self).flags

  @highs.GET
  def _getHighs(self) -> Iterator[KeeFlag]:
    if self.__member_index__:
      for flag in type(self).flags:
        if self.index & (1 << flag.index):
          yield flag
    else:
      yield from ()

  @index.GET
  def _getIndex(self) -> int:
    if self.__member_index__ is None:
      raise MissingVariable(self, '__member_index__', int)
    if isinstance(self.__member_index__, int):
      return self.__member_index__
    raise TypeException('__member_index__', self.__member_index__, int)

  @name.GET
  def _getName(self) -> str:
    return '_'.join(f.name for f in self.highs) or 'NULL'

  @names.GET
  def _getNames(self, ) -> frozenset:
    return frozenset(f.name for f in self.highs)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __iter__(self) -> Iterator[Self]:
    yield from self.highs

  def __str__(self, ) -> str:
    infoSpec = """%s.%s [%d]"""
    clsName = type(self).__name__
    info = infoSpec % (clsName, self.name, self.index)
    return textFmt(info)

  __repr__ = __str__

  def __eq__(self, other: Any) -> bool:
    if not isinstance(type(other), KeeFlagsMeta):
      return NotImplemented
    if self.__field_owner__ == other.__field_owner__:
      return True if self.index == other.index else False
    return False

  def __hash__(self, ) -> int:
    return hash((hash(type(self)), *self.highs))

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, index: int) -> None:
    self.__member_index__ = index
